"""
Admin Global api functions as a mixin
"""
from jaseci.api.interface import interface
from jaseci.actor.sentinel import sentinel
import uuid


class global_api():
    """
    Admin global APIs
    """

    @interface.admin_api(cli_args=['name'])
    def global_set(self, name: str, value: str):
        """
        Set a global
        """
        ret = {'success': True}
        if(name == 'GLOB_SENTINEL' or name in self._valid_configs):
            ret['response'] = f"{name} is sacred!"
            ret['success'] = False
        else:
            self._h.save_glob(name, value)
            ret['response'] = f"Global variable '{name}' to '{value}' set!"
        return ret

    @interface.admin_api(cli_args=['name'])
    def global_delete(self, name: str):
        """
        Delete a global
        """
        ret = {'success': True}
        if(name == 'GLOB_SENTINEL' or name in self._valid_configs):
            ret['response'] = f"{name} is sacred!"
            ret['success'] = False
        else:
            self._h.destroy_glob(name)
            ret['response'] = f"Global {name} deleted."
        return ret

    @interface.admin_api()
    def global_sentinel_set(self, snt: sentinel = None):
        """
        Set sentinel as globally accessible
        """
        for i in snt.get_deep_obj_list():
            i.make_read_only()
        self._h.save_glob('GLOB_SENTINEL', snt.jid)
        return {'response': f"Global sentinel set to '{snt}'!"}

    @interface.admin_api()
    def global_sentinel_unset(self):
        """
        Set sentinel as globally accessible
        """
        current = self.global_get('GLOB_SENTINEL')['value']
        if(current):
            snt = self._h.get_obj(self._m_id, uuid.UUID(current))
            for i in snt.get_deep_obj_list():
                i.make_private()
        self._h.destroy_glob('GLOB_SENTINEL')
        return {'response': f"Global sentinel cleared!"}



from enum import Enum, auto


class op(Enum):
    # NOOP = auto()
    IDS_CLEAR = auto()  # [ref id_list]
    IDS_ADD_OBJ = auto()  # [ref ids, ref obj]
    IDS_GET_LEN = auto()  # [ref dest, ref ids]
    PUSH_SCOPE_W = auto()  # []
    POP_SCOPE = auto()  # []
    SET_LIVE_VAR = auto()  # [str dest, ref value, md_index]
    SET_REF_VAR = auto()  # [ref dest, ref src]
    SET_REF_VARI = auto()  # [ref dest, imm value]
    CREATE_CTX_VAR = auto()  # [ref obj, str dest, ref val, bool is_private]
    SET_ANCHOR = auto()  # [ref obj, str name]
    B_NEQ = auto()  # [ref src1, ref src2, label]
    B_NEQI = auto()  # [ref src1, imm val, label]
    B_NIT = auto()  # [ref src1, typ, label]
    B_A = auto()  # [label]
    PLUS = auto()  # [dest, src, src]
    END = auto()   # []



"""
AST for Jac
"""
from antlr4 import InputStream, CommonTokenStream, ParseTreeListener
from antlr4.error.ErrorListener import ErrorListener

from jaseci.utils.utils import logger, parse_str_token
from jaseci.jac.jac_parse.jacLexer import jacLexer
from jaseci.jac.jac_parse.jacParser import jacParser, ParseTreeWalker
import os


class ast():
    """
    AST Nodes

    The kind field is used to represent the grammar rule
    TODO: Error handling if jac program has errors
    """

    _ast_head_map = {}

    def __init__(self, mod_name, mod_dir='./', jac_text=None,
                 parse_errors=None, start_rule='start', fresh_start=True):
        if(fresh_start):
            ast._ast_head_map = {}
        self.name = 'unparsed'
        self.kind = 'unparsed'
        self.context = {}
        self._parse_errors = parse_errors if parse_errors else []
        self._start_rule = start_rule
        self.mod_name = mod_name if mod_name is not None else "@default"
        self._mod_dir = mod_dir
        self.line = 0
        self.column = 0
        self.kid = []
        if(jac_text):
            self.parse_jac_str(jac_text)

    def is_rule(self):
        """Returns true if node is a rule"""
        return self.kind == 'rule'

    def is_terminal(self):
        """Returns true if node is a terminal"""
        return self.kind == 'terminal'

    def token(self):
        if(not self.is_terminal()):
            logger.error(
                str(f"Non terminals (rules) don't have token info - {self}")
            )
            return None
        else:
            return self.context["token"]

    def token_text(self):
        if(self.is_terminal()):
            return self.token()['text']

    def token_symbol(self):
        if(self.is_terminal()):
            return self.token()['symbol']

    def parse_jac_str(self, jac_str):
        """Parse language and build ast from string"""
        ast._ast_head_map[self.mod_name] = self
        input_stream = InputStream(jac_str)
        lexer = jacLexer(input_stream)
        stream = CommonTokenStream(lexer)
        errors = self.jac_tree_error(self)
        parser = jacParser(stream)
        parser.removeErrorListeners()
        parser.addErrorListener(errors)
        tree = getattr(parser, self._start_rule)()
        builder = self.jac_tree_builder(self)
        walker = ParseTreeWalker()
        walker.walk(builder, tree)

        if(self._parse_errors):
            logger.error(
                str(f"Parse errors encountered - {self}")
            )

    def __str__(self):
        res = f'{self.kind}:{self.name}'
        if(self.is_terminal()):
            res += \
                f':{self.context["token"]["text"]}'
        return res

    def __repr__(self):
        return self.__str__()

    def get_tokens(self):
        """Return list of all tokens derived from this ast node"""
        tokens = []
        if('token' in self.context):
            tokens.append(self.context['token'])
            return tokens
        else:
            for i in self.kid:
                tokens = tokens + i.get_tokens()
        return tokens

    def get_text(self):
        """Get source text for this node"""
        ret = ''
        for i in self.get_tokens():
            ret += f"{i['text']} "
        return ret

    class jac_tree_builder(ParseTreeListener):
        """Converter class from Antlr trees to Jaseci Tree"""

        def __init__(self, tree_root):
            self.tree_root = tree_root
            self.node_stack = []

        def run_import_module(self, jac_ast):
            """
            import_module:
                KW_IMPORT LBRACE (import_items | '*') RBRACE
                KW_WITH STRING SEMI;

            TODO: Check for duplicate imports and ignore if imported
            """
            kid = jac_ast.kid
            fn = os.path.join(self.tree_root._mod_dir, parse_str_token(
                kid[-2].token_text()))
            mod_name = os.path.basename(fn)
            from_mod = self.tree_root.mod_name
            logger.debug(f"Importing items from {mod_name} to {from_mod}...")
            parsed_ast = None
            if(mod_name in ast._ast_head_map.keys()):
                parsed_ast = ast._ast_head_map[mod_name]
            elif(os.path.isfile(fn)):
                with open(fn, 'r') as file:
                    jac_text = file.read()
                mdir = os.path.dirname(os.path.realpath(fn))+'/'
                parsed_ast = ast(jac_text=jac_text, mod_name=mod_name,
                                 mod_dir=mdir, fresh_start=False)
            else:
                err = f"Module not found for import! {mod_name} from" +\
                    f" {from_mod}"
                self.tree_root._parse_errors.append(err)
            if(parsed_ast):
                self.tree_root._parse_errors += parsed_ast._parse_errors
                import_elements = list(filter(lambda x:
                                              x.name == 'element',
                                              parsed_ast.kid))
                if(kid[2].name == 'STAR_MUL'):
                    return import_elements
                else:
                    return self.run_import_items(
                        kid[2], import_elements)
            else:
                return []

        def run_import_items(self, jac_ast, import_elements):
            """
            import_items:
                KW_WALKER (STAR_MUL | import_names) (COMMA import_items)?
                | KW_NODE (STAR_MUL | import_names) (COMMA import_items)?
                | KW_EDGE (STAR_MUL | import_names) (COMMA import_items)?
                | KW_GRAPH (STAR_MUL | import_names) (COMMA import_items)?;
            """
            kid = jac_ast.kid
            ret_elements = list(filter(lambda x:
                                       x.kid[0].kid[0].name == kid[0].name,
                                       import_elements))
            if(kid[1].name == "import_names"):
                ret_elements = list(filter(lambda x:
                                           x.kid[0].kid[1].token_text() in
                                           self.run_import_names(kid[1]),
                                           ret_elements))
            if(kid[-1].name == "import_items"):
                return ret_elements + self.run_import_items(kid[-1],
                                                            import_elements)

            return ret_elements

        def run_import_names(self, jac_ast):
            """
            import_names:
                DBL_COLON NAME
                | DBL_COLON LBRACE name_list RBRACE;
            """
            kid = jac_ast.kid
            if(kid[1].name == "NAME"):
                return [kid[1].token_text()]
            else:
                return self.run_name_list(kid[2])

        def run_name_list(self, jac_ast):
            """
            name_list: NAME (COMMA NAME)*;
            """
            kid = jac_ast.kid
            ret = []
            for i in kid:
                if(i.name == 'NAME'):
                    ret.append(i.token_text())
            return ret

        def enterEveryRule(self, ctx):
            """Visits every node in antlr parse tree"""
            if(len(self.node_stack) == 0):
                new_node = self.tree_root
            else:
                new_node = ast(mod_name=self.tree_root.mod_name,
                               mod_dir=self.tree_root._mod_dir,
                               fresh_start=False)
            new_node.name = jacParser.ruleNames[ctx.getRuleIndex()]
            new_node.kind = 'rule'
            new_node._parse_errors = self.tree_root._parse_errors
            new_node.line = ctx.start.line
            new_node.column = ctx.start.column

            if(len(self.node_stack) and new_node.name != 'import_module'):
                self.node_stack[-1].kid.append(new_node)
            self.node_stack.append(new_node)

        def exitEveryRule(self, ctx):
            """Overloaded function that visits every node on exit"""
            top = self.node_stack.pop()
            if(top.name == 'import_module'):
                self.node_stack[-1].kid += self.run_import_module(top)

        def visitTerminal(self, node):
            """Visits terminals as walker walks, adds ast node"""
            new_node = ast(mod_name=self.tree_root.mod_name,
                           mod_dir=self.tree_root._mod_dir, fresh_start=False)
            new_node.name = jacParser.symbolicNames[node.getSymbol().type]
            new_node.kind = 'terminal'
            new_node.line = node.getSymbol().line
            new_node.column = node.getSymbol().column

            token = {
                'symbol':   jacParser.symbolicNames[node.getSymbol().type],
                'text':     node.getSymbol().text,
            }
            new_node.context['token'] = token

            self.node_stack[-1].kid.append(new_node)

    class jac_tree_error(ErrorListener):
        """Accumulate errors as parse tree is walked"""

        def __init__(self, tree_root):
            self.tree_root = tree_root

        def syntaxError(self, recognizer, offendingSymbol,
                        line, column, msg, e):
            """Add error to error list"""
            self.tree_root._parse_errors.append(
                f"{str(self.tree_root.mod_name)}: line {str(line)}:"
                f"{str(column)} - {self.tree_root} - {msg}"
            )



"""
Super master handler for each user of Jaseci, serves as main interface between
between user and Jaseci
"""

from jaseci.element.master import master
from jaseci.api.logger_api import logger_api
from jaseci.api.config_api import config_api
from jaseci.api.global_api import global_api
from jaseci.api.super_api import super_api
from jaseci.api.stripe_api import stripe_api
from jaseci.api.actions_api import actions_api


class super_master(master, logger_api, actions_api, config_api, global_api,
                   super_api, stripe_api):
    """Master with admin APIs"""

    def __init__(self, *args, **kwargs):
        master.__init__(self, *args, **kwargs)
        stripe_api.__init__(self)



from jaseci.utils.utils import TestCaseHelper
from unittest import TestCase
from jaseci.graph.node import node
from jaseci.attr import action
from jaseci.graph.edge import edge
from jaseci.element import element
from jaseci.utils.mem_hook import mem_hook
from jaseci.actor.architype import architype


class node_tests(TestCaseHelper, TestCase):
    """Tests for the funcationality of Jaseci node class"""

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_node_connections(self):
        """Test connecting and disconnecting etc of nodes"""
        node1 = node(m_id='anon', h=mem_hook())
        node2 = node(m_id='anon', h=node1._h)
        node3 = node(m_id='anon', h=node1._h)
        node4 = node(m_id='anon', h=node1._h)
        node1.attach_outbound(node2)
        node3.attach_outbound(node2)
        node1.attach_inbound(node2)
        node3.attach_inbound(node2)
        self.assertEqual(len(node2.inbound_nodes()), 2)
        self.assertEqual(node2.inbound_nodes()[0], node1)
        self.assertTrue(node2.is_attached_out(node3))
        self.assertTrue(node2.is_attached_in(node1))
        self.assertFalse(node4.is_equivalent(node2))

        node2.detach_inbound(node1)
        node2.detach_inbound(node3)
        node2.detach_outbound(node1)
        node2.detach_outbound(node3)

        self.assertTrue(node4.is_equivalent(node2))
        self.assertTrue(node3.is_equivalent(node1))
        self.assertTrue(node3.is_equivalent(node2))
        self.assertEqual(len(node1.inbound_nodes()), 0)
        self.assertEqual(len(node2.inbound_nodes()), 0)
        self.assertEqual(len(node3.inbound_nodes()), 0)
        self.assertEqual(len(node4.inbound_nodes()), 0)

    def test_add_context_to_node_and_destroy(self):
        """Test adding and removing contexts nodes"""
        node1 = node(m_id='anon', h=mem_hook())
        node1.context["yeah dude"] = "SUP"
        self.assertEqual(
            node1.context["yeah dude"], 'SUP'
        )
        self.assertEqual(len(node1.context.keys()), 1)
        self.assertTrue('yeah dude' in node1.context.keys())
        self.assertFalse('yeah  dude' in node1.context.keys())

    def test_add_entry_action_to_node_and_destroy(self):
        """Test connecting and disconnecting etc of nodes"""
        node1 = architype(m_id='anon', h=mem_hook())
        act = action.action(m_id='anon', h=node1._h,
                            name="yeah dude", value="SUP")
        node1.entry_action_ids.add_obj(act)
        self.assertEqual(
            node1.entry_action_ids.get_obj_by_name("yeah dude").value, 'SUP'
        )
        self.assertEqual(len(node1.entry_action_ids), 1)
        self.assertTrue(node1.entry_action_ids.has_obj_by_name('yeah dude'))
        self.assertFalse(node1.entry_action_ids.has_obj_by_name('yeah  dude'))

        node1.entry_action_ids.destroy_obj_by_name(name="yeah dude")
        self.assertEqual(len(node1.entry_action_ids), 0)
        self.assertFalse(node1.entry_action_ids.has_obj_by_name('yeah dude'))

    def test_adding_and_removing_from_hdnodes(self):
        """Test adding nodes and removing them from HDGDs"""
        node1 = node(m_id='anon', h=mem_hook())
        hdgd1 = node(m_id='anon', h=node1._h, name="yeah dude", dimension=1)
        node1.make_member_of(hdgd1)
        self.assertEqual(
            node1.parent_node_ids.obj_list()[0], hdgd1
        )
        self.assertEqual(
            hdgd1.member_node_ids.obj_list()[0], node1
        )

        hdgd2 = node(m_id='anon', h=node1._h, dimension=2)
        node1.make_member_of(hdgd2)
        self.assertNotIn(hdgd2.id, node1.parent_node_ids)

        node1.leave_memebership_of(hdgd1)
        self.assertEqual(len(hdgd1.member_node_ids), 0)

    def test_dimensions_must_match_to_connect_nodes(self):
        """Test dimension matching requirement for node connection"""
        node1 = node(m_id='anon', h=mem_hook())
        hdgd1 = node(m_id='anon', h=node1._h, name="yeah dude", dimension=1)
        node1.attach_outbound(hdgd1)
        self.assertFalse(node1.is_attached_out(hdgd1))

    def test_inherit_from_element_edge(self):
        """Test that inheriting params with kwargs works"""
        hook = element.mem_hook()
        a = edge(name="my edge", m_id='anon', h=hook,
                      to_node=node(m_id='anon', h=hook),
                      from_node=node(m_id='anon', h=hook))
        self.assertEqual(a.name, "my edge")



"""
Sentinel api functions as a mixin
"""
from jaseci.api.interface import interface
from jaseci.utils.id_list import id_list
from jaseci.actor.sentinel import sentinel
from jaseci.utils.utils import b64decode_str
import uuid


class sentinel_api():
    """
    Sentinel APIs
    """

    def __init__(self):
        self.active_snt_id = None
        self.sentinel_ids = id_list(self)

    @interface.private_api(cli_args=['code'])
    def sentinel_register(self, name: str = 'default', code: str = '',
                          code_dir: str = './',
                          encoded: bool = False, auto_run: str = 'init',
                          auto_gen_graph: bool = True, ctx: dict = {},
                          set_active: bool = True):
        """
        Create blank or code loaded sentinel and return object
        Auto_run is the walker to execute on register (assumes active graph
        is selected)
        """
        snt = self.sentinel_ids.get_obj_by_name(name, silent=True)
        new_gph = None
        if(not snt):
            snt = sentinel(m_id=self._m_id, h=self._h, name=name)
            self.sentinel_ids.add_obj(snt)
            if(auto_gen_graph):
                new_gph = self.graph_create(set_active=True)
        if(code):
            if (encoded):
                code = b64decode_str(code)
            snt.register_code(code, code_dir)
            if(not snt.is_active):
                return {'response': 'Error in jac code',
                        'errors': snt.errors,
                        'success': False}
        if(snt.walker_ids.has_obj_by_name(auto_run) and self.active_gph_id):
            nd = self._h.get_obj(self._m_id, uuid.UUID(self.active_gph_id))
            self.walker_run(name=auto_run, nd=nd, ctx=ctx,
                            snt=snt)
        if(set_active):
            self.sentinel_active_set(snt)
        self.extract_snt_aliases(snt)
        if(new_gph):
            return [snt.serialize(), new_gph]
        return [snt.serialize()]

    @interface.private_api()
    def sentinel_pull(self, set_active: bool = True,
                      on_demand: bool = True):
        """
        Copies global sentinel to local master
        """
        glob_id = self._h.get_glob('GLOB_SENTINEL')
        if(not glob_id):
            return {'response': 'No global sentinel is available!',
                    'success': False}
        g_snt = self._h.get_obj(self._m_id, uuid.UUID(glob_id)).duplicate()

        snt = self.sentinel_ids.get_obj_by_name(g_snt.name, silent=True)
        if(not snt):
            snt = sentinel(m_id=self._m_id, h=self._h, name=g_snt.name)
            self.sentinel_ids.add_obj(snt)
        elif(on_demand and snt.is_active):
            return {'response': f'{snt} already active!',
                    'success': True}
        if(set_active):
            self.sentinel_active_set(snt)
        return self.sentinel_set(code=g_snt.code_ir, snt=snt,
                                 mode='ir')

    @interface.private_api()
    def sentinel_get(self, snt: sentinel = None,
                     mode: str = 'default', detailed: bool = False):
        """
        Get a sentinel rendered with specific mode
        Valid modes: {default, code, ir, }
        """
        if(mode == 'code'):
            return snt._jac_ast.get_text()
        elif(mode == 'ir'):
            return snt.ir_dict()
        else:
            return snt.serialize(detailed=detailed)

    @interface.private_api(cli_args=['code'])
    def sentinel_set(self, code: str, code_dir: str = './',
                     encoded: bool = False,
                     snt: sentinel = None, mode: str = 'default'):
        """
        Set code/ir for a sentinel, only replaces walkers/archs in sentinel
        Valid modes: {code, ir, }
        """
        if (encoded):
            code = b64decode_str(code)
        if(mode == 'code' or mode == 'default'):
            snt.register_code(code, code_dir)
        elif(mode == 'ir'):
            snt.apply_ir(code)
            snt.run_start(snt._jac_ast)
            if(snt.runtime_errors):
                snt.is_active = False
        else:
            return {'response': f'Invalid mode to set {snt}',
                    'success': False}
        if(snt.is_active):
            self.extract_snt_aliases(snt)
            return {'response': f'{snt} registered and active!',
                    'success': True}
        else:
            return {'response': f'{snt} code issues encountered!',
                    'success': False,
                    'errors': snt.errors+snt.runtime_errors}

    @interface.private_api()
    def sentinel_list(self, detailed: bool = False):
        """
        Provide complete list of all sentinel objects
        """
        snts = []
        for i in self.sentinel_ids.obj_list():
            snts.append(i.serialize(detailed=detailed))
        return snts

    @interface.private_api()
    def sentinel_test(self, snt: sentinel = None,
                      detailed: bool = False):
        """
        Run battery of test cases within sentinel and provide result
        """
        return snt.run_tests(detailed=detailed)

    @interface.private_api(cli_args=['snt'])
    def sentinel_active_set(self, snt: sentinel):
        """
        Sets the default sentinel master should use
        """
        self.active_snt_id = snt.jid
        self.alias_register('active:sentinel', snt.jid)
        return [f'Sentinel {snt.id} set as default']

    @interface.private_api()
    def sentinel_active_unset(self):
        """
        Unsets the default sentinel master should use
        """
        self.active_snt_id = None
        self.alias_delete('active:sentinel')
        return ['Default sentinel unset']

    @interface.private_api()
    def sentinel_active_global(self, detailed: bool = False):
        """
        Sets the default master sentinel to the global sentinel
        Exclusive OR with pull strategy
        """
        ret = {'success': False, 'sentinel': None}
        glob_id = self._h.get_glob('GLOB_SENTINEL')
        if(not glob_id):
            ret['response'] = 'No global sentinel is available!'
        else:
            self.active_snt_id = "global"  # Resolved in interface
            self.alias_register('active:sentinel', glob_id)
            ret['sentinel'] = self._h.get_obj(
                self._m_id, uuid.UUID(glob_id)).serialize(detailed=detailed)
            ret['success'] = True
            ret['response'] = f'Global sentinel {glob_id} set as default'
        return ret

    @interface.private_api()
    def sentinel_active_get(self, detailed: bool = False):
        """
        Returns the default sentinel master is using
        """
        id = self.active_snt_id
        if(id == 'global'):
            id = self._h.get_glob('GLOB_SENTINEL')
        if(not id):
            return {'response': 'No default sentinel is selected!'}
        else:
            default = self._h.get_obj(
                self._m_id, uuid.UUID(id))
            return default.serialize(detailed=detailed)

    @interface.private_api(cli_args=['snt'])
    def sentinel_delete(self, snt: sentinel):
        """
        Permanently delete sentinel with given id
        """
        self.remove_snt_aliases(snt)
        if(self.active_snt_id == snt.jid):
            self.sentinel_active_unset()
        self.sentinel_ids.destroy_obj(snt)
        return [f'Sentinel {snt.id} successfully deleted']

    def active_snt(self):
        return self._h.get_obj(
            self._m_id, uuid.UUID(self.active_snt_id))

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.sentinel_ids.obj_list():
            i.destroy()



"""
Sentinel interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.graph.node import node
from jaseci.graph.edge import edge
from jaseci.jac.interpreter.interp import interp
from jaseci.jac.machine.jac_scope import jac_scope
from jaseci.utils.utils import parse_str_token
from jaseci.jac.machine.jac_value import jac_elem_unwrap as jeu


class architype_interp(interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    def run_architype(self, jac_ast):
        """
        architype:
            KW_NODE NAME (COLON INT)? attr_block
            | KW_EDGE NAME attr_block
            | KW_GRAPH NAME graph_block;
        """
        if(jac_ast is None):  # Using defaults
            if(self.kind == 'node' and self.name in ['root', 'generic']):
                return node(m_id=self._m_id, h=self._h,
                            kind=self.kind, name=self.name)
            elif(self.kind == 'edge' and self.name in ['generic']):
                return edge(m_id=self._m_id, h=self._h,
                            kind=self.kind, name=self.name)
        kid = self.set_cur_ast(jac_ast)
        self.push_scope(
            jac_scope(
                parent=self,
                has_obj=self,
                action_sets=[]))
        if(kid[0].name == 'KW_NODE'):
            item = node(m_id=self._m_id, h=self._h,
                        kind=kid[0].token_text(), name=kid[1].token_text())
            if(kid[2].name == 'COLON'):
                item.dimension = int(kid[3].token_text())
            self.run_attr_block(kid[-1], item)
        elif(kid[0].name == 'KW_EDGE'):
            item = edge(m_id=self._m_id, h=self._h,
                        kind=kid[0].token_text(), name=kid[1].token_text())
            self.run_attr_block(kid[-1], item)
        elif (kid[0].name == 'KW_GRAPH'):
            item = self.run_graph_block(kid[-1])
        elif (jac_ast.name == 'graph_block'):  # used in jac tests
            item = self.run_graph_block(jac_ast)
        item.parent_id = self.parent().id
        self.pop_scope()
        return item

    def run_attr_block(self, jac_ast, obj):
        """
        attr_block:
            LBRACE (attr_stmt)* RBRACE
            | COLON (attr_stmt)* SEMI
            | SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if(i.name == 'attr_stmt'):
                self.run_attr_stmt(i, obj)
        self._can_compiled_flag = True

    def run_graph_block(self, jac_ast):
        """
        graph_block: graph_block_spawn | graph_block_dot;
        """
        kid = self.set_cur_ast(jac_ast)
        return getattr(self, f'run_{kid[0].name}')(kid[0])

    def run_graph_block_spawn(self, jac_ast):
        """
        graph_block_spawn:
            LBRACE has_root KW_SPAWN code_block RBRACE
            | COLON has_root KW_SPAWN code_block SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        root_name = self.run_has_root(kid[1])
        m = interp(parent_override=self.parent(), caller=self)
        m.push_scope(jac_scope(parent=self,
                               has_obj=None,
                               action_sets=[]))
        m.run_code_block(kid[3])
        local_state = m._jac_scope.local_scope
        self.report = self.report + m.report
        if(root_name in local_state.keys()):
            obj = jeu(local_state[root_name], parent=self)
            if(not isinstance(obj, node)):
                self.rt_error(f"{root_name} is {type(obj)} not node!",
                              kid[2])
            return obj
        else:
            self.rt_error(f"Graph didn't produce root node!",
                          kid[2])
            return None

    def run_graph_block_dot(self, jac_ast):
        """
        graph_block_dot:
            LBRACE has_root dot_graph RBRACE
            | COLON has_root dot_graph SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        graph_state = {
            'strict': False,
            'digraph': True,
            'node_ops': [],
            'edge_ops': []
        }
        root_name = self.run_has_root(kid[1])
        self.run_dot_graph(kid[2], graph_state)

        nodes_def = {}
        for op in graph_state['node_ops']:
            node_id = op.pop('id', None)
            # TODO: is op actually needed?
            op.pop('op')
            if (node_id not in nodes_def):
                nodes_def[node_id] = op
            nodes_def[node_id].update(op)

        if (root_name not in nodes_def):
            del nodes_def
            self.rt_error(f"Graph didn't produce root node!",
                          kid[1])
            return None

        # Create node objects
        node_objs = {}
        for node_id, node_def in nodes_def.items():
            node_name = node_def.pop('node', None)
            if(node_name is None):
                self.rt_error('Missing "node" attribute for node.')
                continue
            node_obj = self.parent().run_architype(
                node_name, kind='node', caller=self)
            node_obj.set_context(node_def)

            # Overwrite node name with _n_name_ in the attrs if defined
            node_objs[node_id] = node_obj

        # Create edge objects
        edge_objs = []
        for op in graph_state['edge_ops']:
            edge_kind = op.pop('edge', None)
            if(edge_kind):
                edge_obj = self.parent().run_architype(
                    edge_kind, kind='edge', caller=self)
            else:
                edge_obj = edge(m_id=self._m_id, h=self._h,
                                kind='edge', name='generic')

            lhs_node_id = op.pop('lhs_node_id')
            rhs_node_id = op.pop('rhs_node_id')
            op.pop('op')
            op.pop('is_directional')
            edge_obj.set_context(op)
            lhs_node = node_objs.get(lhs_node_id, None)
            if(lhs_node is None):
                del nodes_def
                del node_objs
                del edge_objs
                self.rt_error('Invalid from node for edge')
                return None
            rhs_node = node_objs.get(rhs_node_id, None)
            if(rhs_node is None):
                del nodes_def
                del node_objs
                del edge_objs
                self.rt_error('Invalid to node for edge')

            lhs_node.attach_outbound(rhs_node, [edge_obj])
            edge_objs.append(edge_obj)

            # TODO: handle non-directional edge once that's supported in jac

        return node_objs[root_name]

    def run_has_root(self, jac_ast):
        """
        has_root: KW_HAS KW_ANCHOR NAME SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        return kid[2].token_text()

    def run_dot_graph(self, jac_ast, graph_state):
        """
        dot_graph:
            KW_STRICT? (KW_GRAPH | KW_DIGRAPH) dot_id? '{' dot_stmt_list '}';
        """
        # TODO: jac should support multiple edges between the same node but it
        # currently does not. once that's updates, need to update here
        # accordingly. We will only support non strict graph.
        kid = self.set_cur_ast(jac_ast)
        if (kid[0].name == 'KW_STRICT'):
            graph_state['strict'] = True
            kid = kid[1:]
        if (kid[0].name == 'KW_GRAPH'):
            graph_state['digraph'] = False
        kid = kid[1:]
        if (kid[0].name == 'dot_id'):
            kid = kid[1:]
        self.run_dot_stmt_list(kid[1], graph_state)

    def run_dot_stmt_list(self, jac_ast, graph_state):
        """
        dot_stmt_list: (dot_stmt ';'?)*
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if(i.name == 'dot_stmt'):
                self.run_dot_stmt(i, graph_state)

    def run_dot_stmt(self, jac_ast, graph_state):
        """
        dot_stmt:
            dot_node_stmt
            | dot_edge_stmt
            | dot_attr_stmt
            | dot_id '=' dot_id
            | dot_subgraph
        """
        kid = self.set_cur_ast(jac_ast)
        if (kid[0] == 'dot_id'):
            pass
        else:
            getattr(self, f'run_{kid[0].name}')(kid[0], graph_state)

    def run_dot_attr_stmt(self, jac_ast, graph_state):
        """
        dot_attr_stmt: (KW_GRAPH | KW_NODE | KW_EDGE) dot_attr_list
        """
        pass

    def run_dot_attr_list(self, jac_ast):
        """
        dot_attr_list: ('[' dot_a_list? ']')+
        """
        kid = self.set_cur_ast(jac_ast)
        attrs = {}
        while (len(kid) > 0):
            attrs.update(self.run_dot_a_list(kid[1]))
            kid = kid[3:]
        return attrs

    def run_dot_a_list(self, jac_ast):
        """
        dot_a_list: (dot_id('=' dot_id)? ','?)+
        """
        kid = self.set_cur_ast(jac_ast)
        a_list = {}
        while (len(kid) > 0):
            lhs_id = self.run_dot_id(kid[0])
            kid = kid[1:]
            if (len(kid) == 0 or kid[0].token_text() != '='):
                # If there is no rhs, treat it as a boolean value of True
                # e.g. [is_active, color=red] sets the attributes
                # is_active as True and color as "red"
                rhs_id = True
            else:
                rhs_id = self.run_dot_id(kid[1])
                kid = kid[2:]

            a_list[lhs_id] = rhs_id

            # deal with the optional comma
            if (len(kid) > 0 and kid[0].token_text() == ','):
                kid = kid[1:]

        return a_list

    def run_dot_edge_stmt(self, jac_ast, graph_state):
        """
        dot_edge_stmt: (dot_node_id | dot_subgraph) dot_edgeRHS dot_attr_list?
        """
        kid = self.set_cur_ast(jac_ast)
        if (kid[0] == 'dot_subgraph'):
            self.rt_error('Subgraphs not supported!', kid[0])
            return
        lhs_id = str(self.run_dot_node_id(kid[0]))
        graph_state['node_ops'].append({
            'op': 'create',
            'id': lhs_id
        })
        edge_attrs = {}
        if(kid[-1].name == 'dot_attr_list'):
            edge_attrs = self.run_dot_attr_list(kid[-1])
        self.run_dot_edgeRHS(kid[1], graph_state, lhs_id, edge_attrs)

    def run_dot_edgeRHS(self, jac_ast, graph_state, lhs_id, edge_attrs):
        """
        dot_edgeRHS: (dot_edgeop(dot_node_id | dot_subgraph))+
        """
        kid = self.set_cur_ast(jac_ast)
        cur_lhs_id = lhs_id
        while (len(kid) > 0):
            is_directional = self.run_dot_edgeop(kid[0])
            if (kid[1] == 'dot_subgraph'):
                self.rt_error('Subgraphs not supported!', kid[1])
                return
            rhs_id = str(self.run_dot_node_id(kid[1]))
            # Add create node rhs
            graph_state['node_ops'].append({
                'op': 'create',
                'id': rhs_id
            })
            # Add create edge
            edge_op = {
                'op': 'create',
                'is_directional': is_directional,
                'lhs_node_id': cur_lhs_id,
                'rhs_node_id': rhs_id
            }
            edge_op.update(edge_attrs)
            graph_state['edge_ops'].append(edge_op)
            cur_lhs_id = rhs_id
            kid = kid[2:]

    def run_dot_edgeop(self, jac_ast):
        """
        dot_edgeop: '->' | '--'
        """
        kid = self.set_cur_ast(jac_ast)
        if (kid[0].token_text() == '->'):
            return True
        else:
            return False

    def run_dot_node_stmt(self, jac_ast, graph_state):
        """
        dot_node_stmt: dot_node_id dot_attr_list?
        """
        kid = self.set_cur_ast(jac_ast)
        node_id = self.run_dot_node_id(kid[0])
        node_attrs = {}
        if (kid[-1].name == 'dot_attr_list'):
            node_attrs = self.run_dot_attr_list(kid[-1])
        node_create_op = {
            'op': 'create',
            'id': node_id,
        }
        graph_state['node_ops'].append(node_create_op)

        node_update_op = {
            'op': 'update',
            'id': node_id,
        }
        node_update_op.update(node_attrs)
        graph_state['node_ops'].append(node_update_op)

    def run_dot_node_id(self, jac_ast):
        """
        dot_node_id: dot_id dot_port?
        """
        kid = self.set_cur_ast(jac_ast)
        if (kid[-1].name == 'dot_port'):
            self.rt_warn('Node ports not supported')
        return self.run_dot_id(kid[0])

    def run_dot_port(self, jac_ast, graph_state):
        """
        dot_port: ':' dot_id(':' dot_id)?
        """
        pass

    def run_dot_subgraph(self, jac_ast, graph_state):
        """
        dot_subgraph: (KW_SUBGRAPH dot_id?)? '{' dot_stmt_list '}'
        """
        pass

    def run_dot_id(self, jac_ast):
        """
        dot_id:
            NAME
            | STRING
            | INT
            | FLOAT
            | KW_GRAPH
            | KW_NODE
            | KW_EDGE;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'INT'):
            return int(kid[0].token_text())
        elif (kid[0].name == 'FLOAT'):
            return float(kid[0].token_text())
        elif(kid[0].name == 'STRING'):
            return parse_str_token(kid[0].token_text())
        return kid[0].token_text()



from antlr4 import InputStream, CommonTokenStream
from jaseci.jac.jac_parse.jacLexer import jacLexer
from jaseci.jac.jac_parse.jacParser import jacParser
from jaseci.utils.mem_hook import mem_hook
from jaseci.actor.sentinel import sentinel
from jaseci.graph.graph import graph
from jaseci.element.super_master import super_master
from jaseci.element.master import master

from jaseci.utils.utils import TestCaseHelper
from unittest import TestCase
import jaseci.tests.jac_test_code as jtc


class jac_tests(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_antlr4_parsing_lifelogify(self):
        """Basic test of jac grammar with lifelogify program"""
        input_stream = InputStream(jtc.prog1)
        lexer = jacLexer(input_stream)
        stream = CommonTokenStream(lexer)
        parser = jacParser(stream)
        parser.start()
        self.assertEqual(parser.getNumberOfSyntaxErrors(), 0)

    def test_sentinel_loading_jac_code(self):
        """Test the generation of jaseci trees for programs in grammar"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.prog1)
        self.assertIsNotNone(
            sent.walker_ids.get_obj_by_name('get_gen_day'))
        self.assertIsNotNone(
            sent.arch_ids.get_obj_by_name('week', kind='node'))

    def test_sentinel_loading_jac_code_multiple_times(self):
        """Test registering resets correctly for multiple attempts"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.prog0)
        num_walkers = len(sent.walker_ids)
        num_arch = len(sent.arch_ids)
        sent.register_code(jtc.prog0)
        self.assertEqual(len(sent.walker_ids), num_walkers)
        self.assertEqual(len(sent.arch_ids), num_arch)
        sent.register_code(jtc.prog0)
        self.assertEqual(len(sent.walker_ids), num_walkers)
        self.assertEqual(len(sent.arch_ids), num_arch)

    def test_sentinel_register_dep_on_static_errors(self):
        """Test Jac registering is dependant on correct static/dynamic code"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        correct = "node b { has anchor a; }"
        wrong1 = "adfdsf"
        sent.register_code(correct)
        self.assertTrue(sent.is_active)
        sent.register_code(wrong1)
        self.assertFalse(sent.is_active)
        sent.register_code(correct)
        self.assertTrue(sent.is_active)
        sent.register_code(wrong1)
        self.assertFalse(sent.is_active)
        sent.register_code(correct)
        self.assertTrue(sent.is_active)

    def test_sentinel_loading_arhitype(self):
        """Test the generation of jaseci trees for programs in grammar"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.prog1)
        self.assertGreater(
            len(sent.arch_ids.get_obj_by_name('month', kind='node').code_ir),
            5)

    def test_sentinel_running_basic_walker(self):
        """Test the execution of a basic walker building graph"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name(
            'life', kind='node').run()
        test_walker = \
            sent.walker_ids.get_obj_by_name('get_gen_day')
        test_walker.prime(test_node)
        test_walker.context['date'] = '2010-08-03T03:00:00.000000'
        test_walker.run()
        self.assertEqual(len(test_node.outbound_nodes()), 1)
        self.assertEqual(test_node.outbound_nodes()[0].name, 'year')
        self.assertEqual(test_node.outbound_nodes()[0].
                         outbound_nodes()[0].outbound_nodes()[0].name, 'week')

    def test_sentinel_setp_running_walker(self):
        """Test the execution of a basic walker building graph"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name(
            'life', kind='node').run()
        test_walker = \
            sent.walker_ids.get_obj_by_name('get_gen_day')
        test_walker.prime(test_node)
        test_walker.context['date'] = '2010-08-03T03:00:00.000000'
        self.assertEqual(len(test_node.outbound_nodes()), 0)
        next = test_walker.step()
        self.assertEqual(test_node.outbound_nodes()[0], next)
        next = test_walker.step()
        next = test_walker.step()
        self.assertEqual(next.name, 'week')
        next = test_walker.step()
        next = test_walker.step()
        self.assertEqual(test_walker.current_step, 5)
        self.assertTrue(next)
        next = test_walker.step()
        self.assertFalse(next)

    def test_walker_writes_through_no_data_gain_loss_on_contexts(self):
        """
        Test that  no loss or gain of data on second trak on second trek
        """
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name(
            'life', kind='node').run()
        test_walker = \
            sent.walker_ids.get_obj_by_name('get_gen_day')
        test_walker.prime(test_node)
        test_walker.context['date'] = '2010-08-03T03:00:00.000000'
        self.assertEqual(test_walker.context['date'],
                         '2010-08-03T03:00:00.000000')
        test_walker.run()
        after_gen = len(gph._h.mem)
        test_walker.prime(test_node)
        test_walker.context['date'] = '2010-08-03T03:00:00.000000'
        test_walker.run()
        after_track = len(gph._h.mem)
        # Tests that more items were created the first time through
        self.assertEqual(after_gen, after_track)

    def test_arch_ids_generate_node_bound_objects_scalably(self):
        """
        Test that arch_ids in sents bind contents to actual nodes
        scalably (node contexts dont get deleted when arch_ids deleted)
        """
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name('life', kind='node').run()
        test_walker = \
            sent.walker_ids.get_obj_by_name('get_gen_day')
        test_walker.prime(test_node)
        test_walker.context['date'] = '2010-08-03T03:00:00.000000'
        self.assertEqual(test_walker.context['date'],
                         '2010-08-03T03:00:00.000000')
        test_walker.run()

        year_node = test_node.outbound_nodes()[0]
        before_del = year_node.context['year']
        sent.register_code(jtc.prog1)
        after_del = year_node.context['year']
        self.assertEqual(before_del, after_del)

    def test_sent_loads_complex_walker_and_arch(self):
        """Test loading attributes of arch and walkers"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name(
            'testnode', kind='node').run()
        test_walker = \
            sent.walker_ids.get_obj_by_name('testwalk')
        test_walker.prime(test_node)
        test_walker.run()
        self.assertEqual(test_node.context['c'],
                         '43Yeah \n"fools"!')

    def test_availabilty_of_global_functions(self):
        """Test preset function loading"""
        self.assertGreater(len(mem_hook().global_action_list), 10)

    def test_multiple_edged_between_nodes_work(self):
        """Test that multiple edges between the same two nodes are allowed"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 3)
        edge_names = [edges[0].name, edges[1].name, edges[2].name]
        self.assertIn('generic', edge_names)
        self.assertIn('apple', edge_names)
        self.assertIn('banana', edge_names)

    def test_multiple_edged_between_nodes_delete_all(self):
        """Test that multiple edges deleted correctly if delete all"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey2)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 0)

    def test_multiple_edged_between_nodes_delete_all_specific(self):
        """Test that multiple edges deleted correctly if delete all"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey2b)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 1)

    def test_multiple_edged_between_nodes_delete_all_labeled(self):
        """Test that multiple edges deleted correctly if delete all"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey2c)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 3)

    def test_multiple_edged_between_nodes_delete_filtered(self):
        """Test that multiple edges deleted correctly if delete filtered"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey3)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 5)

    def test_generic_can_be_used_to_specify_generic_edges(self):
        """Test that generic edge tag works"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey4)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 2)

    def test_can_disconnect_multi_nodes_simultaneously(self):
        """Test disconnecting mutilpe nodes"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey5)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 2)

    def test_can_connect_multi_nodes_simultaneously(self):
        """Test connecting mutilpe nodes"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey6)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 4)

    def test_can_disconnect_multi_nodes_advanced(self):
        """Test disconnecting mutilpe nodes advanced"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edgey7)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        self.assertEqual(len(edges), 3)

    def test_accessing_edges_basic(self):
        """Test accessing Edges"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.edge_access)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        edges = gph.get_all_edges()
        if(edges[0].name == 'apple'):
            self.assertEqual(edges[0].context['v1'], 7)
            self.assertEqual(edges[1].context['x1'], 8)
        else:
            self.assertEqual(edges[1].context['v1'], 7)
            self.assertEqual(edges[0].context['x1'], 8)

    def test_has_assign(self):
        """Test assignment on definition"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.has_assign)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        nodes = gph.get_all_nodes()
        self.assertEqual(len(nodes), 3)
        num = 0
        for i in nodes:
            if(i.name == 'testnode'):
                self.assertEqual(i.context['a'], 8)
                num += 1
        self.assertEqual(num, 2)

    def test_global_get_set(self):
        """Test assignment on definition"""
        mast = super_master(h=mem_hook())
        gph = graph(m_id=mast.jid, h=mast._h)
        sent = sentinel(m_id=mast.jid, h=gph._h)
        sent.register_code(jtc.set_get_global)
        test_walker = \
            sent.walker_ids.get_obj_by_name('setter')
        test_walker.prime(gph)
        test_walker.run()
        test_walker2 = \
            sent.walker_ids.get_obj_by_name('getter')
        test_walker2.prime(gph)
        test_walker2.run()
        self.assertEqual(test_walker2.context['a'], 59)

    def test_global_set_requires_admin(self):
        """Test assignment on definition"""
        mast = master(h=mem_hook())
        gph = graph(m_id=mast.jid, h=mast._h)
        sent = sentinel(m_id=mast.jid, h=gph._h)
        sent.register_code(jtc.set_get_global)
        test_walker = \
            sent.walker_ids.get_obj_by_name('setter')
        test_walker.prime(gph)
        test_walker.run()
        test_walker2 = \
            sent.walker_ids.get_obj_by_name('getter')
        test_walker2.prime(gph)
        test_walker2.run()
        self.assertEqual(test_walker2.context['a'], None)

    def test_sentinel_version_label(self):
        """Test sentinel version labeling"""
        mast = master(h=mem_hook())
        gph = graph(m_id=mast.jid, h=mast._h)
        sent = sentinel(m_id=mast.jid, h=gph._h)
        sent.register_code(jtc.version_label)
        self.assertEqual(sent.version, 'alpha-1.0')

    def test_visibility_builtins(self):
        """Test builtins to see into nodes and edges"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.visibility_builtins)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(2, len(test_walker.report[0][0].keys()))
        self.assertGreaterEqual(len(test_walker.report[0][1].keys()), 7)
        self.assertLess(len(test_walker.report[0][1].keys()), 14)
        self.assertGreaterEqual(len(test_walker.report[0][2].keys()), 14)

    def test_spawn_ctx_for_edges_nodes(self):
        """Test builtins to see into nodes and edges"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.spawn_ctx_edge_node)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0]['age'], 32)
        self.assertEqual(test_walker.report[1]['meeting_place'], 'college')
        self.assertEqual(test_walker.report[2]['name'], 'Jane')
        self.assertEqual(test_walker.report[3]['kind'], 'sister')

    def test_filter_ctx_for_edges_nodes(self):
        """Test builtins to see into nodes and edges"""
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.filter_ctx_edge_node)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0]['age'], 30)
        self.assertEqual(len(test_walker.report[1]), 0)

    def test_null_handling(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.null_handleing)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0], True)
        self.assertEqual(test_walker.report[1], False)
        self.assertEqual(test_walker.report[2], True)
        self.assertEqual(test_walker.report[3], False)

    def test_bool_type_convert(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.bool_type_convert)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0], True)
        self.assertEqual(True, test_walker.report[1]['name'])

    def test_typecasts(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.typecasts)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0], 7.6)
        self.assertEqual(test_walker.report[1], 7)
        self.assertEqual(test_walker.report[2], "7.6")
        self.assertEqual(test_walker.report[3], True)
        self.assertEqual(test_walker.report[4], 7.0)
        self.assertEqual(test_walker.report[5], "Types comes back correct")

    def test_typecast_error(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.typecasts_error)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertIn('Cannot get edges from 7.6',
                      test_walker.runtime_errors[0])
        self.assertIn('Invalid cast of JAC_TYPE.STR to JAC_TYPE.INT',
                      test_walker.runtime_errors[1])

    def test_filter_on_context(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.filter_on_context)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual({'yo': 'Yeah i said'},
                         test_walker.report[0][0])
        self.assertNotIn("name",
                         test_walker.report[0][1].keys())
        self.assertIn("name",
                      test_walker.report[0][2].keys())

    def test_string_manipulation(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.string_manipulation)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], 't')
        self.assertEqual(rep[1], 'tin')
        self.assertEqual(rep[2], 'sting me ')
        self.assertEqual(rep[3], ' TESTING ME  ')
        self.assertEqual(rep[4], ' testing me  ')
        self.assertEqual(rep[5], ' Testing Me  ')
        self.assertEqual(rep[6], ' testing me  ')
        self.assertEqual(rep[7], ' TeSTING ME  ')
        self.assertEqual(rep[8], False)
        self.assertEqual(rep[9], False)
        self.assertEqual(rep[10], False)
        self.assertEqual(rep[11], False)
        self.assertEqual(rep[12], False)
        self.assertEqual(rep[13], False)
        self.assertEqual(rep[14], False)
        self.assertEqual(rep[15], {'a': 5})
        self.assertEqual(rep[16], 2)
        self.assertEqual(rep[17], 5)
        self.assertEqual(rep[18], ['tEsting', 'me'])
        self.assertEqual(rep[19], [' t', 'sting me  '])
        self.assertEqual(rep[20], False)
        self.assertEqual(rep[21], False)
        self.assertEqual(rep[22], ' tEsting you  ')
        self.assertEqual(rep[23], 'tEsting me')
        self.assertEqual(rep[24], 'Esting me')
        self.assertEqual(rep[25], 'tEsting me  ')
        self.assertEqual(rep[26], 'sting me  ')
        self.assertEqual(rep[27], ' tEsting me')
        self.assertEqual(rep[28], ' tEsting m')
        self.assertEqual(rep[29], True)

    def test_list_manipulation(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.list_manipulation)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], [4])
        self.assertEqual(rep[1], [5])
        self.assertEqual(rep[2], [4, 5, 5])
        self.assertEqual(rep[3], [5, 5, 4])
        self.assertEqual(rep[4], [4, 5, 5])
        self.assertEqual(rep[5], 2)
        self.assertEqual(rep[6], [5, 5, 4, 2])
        self.assertEqual(rep[7], [5, 'apple', 4, 2])
        self.assertEqual(rep[8], 1)
        self.assertEqual(rep[9], [5, 'apple', 4])
        self.assertEqual(rep[10], [])

    def test_dict_manipulation(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.dict_manipulation)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], {'four': 4, 'five': 5})
        self.assertEqual(rep[1], {'four': 5, 'five': 5})
        self.assertEqual(rep[2], [['four', 4], ['five', 5]])
        self.assertEqual(rep[3], ['four', 'five'])
        self.assertEqual(rep[4], {'four': 4})
        self.assertEqual(rep[5], [4])
        self.assertEqual(rep[6], {'four': 7})
        self.assertEqual(rep[7], {})

    def test_string_join(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.string_join)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], 'test_me_now')

    def test_sub_list(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.sub_list)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0][0], 5)
        self.assertEqual(rep[0][1], 6)
        self.assertEqual(rep[0][2], 7)

    def test_destroy_and_misc(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.destroy_and_misc)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], 'Josh')
        self.assertEqual(
            rep[1],  {'name': 'Josh', 'age': 32,
                      'birthday': None, 'profession': None})
        self.assertEqual(
            rep[2], {'age': 32, 'birthday': None,
                     'name': 'pete', 'profession': None})
        self.assertEqual(rep[3], [1, 3])
        self.assertEqual(rep[4], {'a': 'b'})
        self.assertEqual(rep[5], [1, 2, 5, 6, 7, 8, 9])
        self.assertEqual(rep[6], [1, 2, 45, 33, 7, 8, 9])
        self.assertEqual(rep[7], None)
        self.assertEqual(rep[8], {'age': 32,
                                  'birthday': None,
                                  'name': 'pete',
                                  'profession': None})
        self.assertEqual(rep[9], True)

    def test_arbitrary_assign_on_element(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.arbitrary_assign_on_element)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0],  {'name': None,
                                   'age': None,
                                   'birthday': None,
                                   'profession': None})

    def test_try_else_stmts(self):
        self.logger_on()
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.try_else_stmts)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], {'args': ('division by zero',),
                                  'col': 8,
                                  'line': 5,
                                  'mod': 'basic',
                                  'msg': 'division by zero',
                                  'type': 'ZeroDivisionError'})
        self.assertEqual(rep[1], 'dont need err')
        self.assertEqual(rep[2], None)
        self.assertEqual(rep[3], 2)

    def test_node_edge_same_name(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.node_edge_same_name)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], {'meeting_place': 'college'})
        self.assertEqual(
            rep[1], {'age': 32, 'birthday': None,
                     'name': 'Josh', 'profession': None})

    def test_testcases(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.testcases)
        sent.run_tests(silent=True)
        self.assertEqual(len(sent.testcases), 4)
        for i in sent.testcases:
            self.assertEqual(i['passed'], True)

    def test_testcase_asserts(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.testcase_asserts)
        sent.run_tests(silent=True)
        self.assertEqual(len(sent.testcases), 3)
        self.assertEqual(sent.testcases[0]['passed'], True)
        self.assertEqual(sent.testcases[1]['passed'], False)
        self.assertEqual(sent.testcases[2]['passed'], False)

    def test_report_not_to_jacset(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.report_not_to_jacset)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertIn('context', rep[0][0].keys())
        self.assertIn('j_type', rep[0][0].keys())

    def test_walker_spawn_unwrap_check(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.walker_spawn_unwrap_check)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertTrue(rep[0].startswith('urn:uuid'))

    def test_std_get_report(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.std_get_report)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep, [3, 5, 6, 7, [3, 5, 6, 7], 8])

    def test_func_with_array_index(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.func_with_array_index)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep, [3, 5, 3])

    def test_rt_error_test1(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.rt_error_test1)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        self.assertIn("List index out of range",
                      test_walker.runtime_errors[0])
        self.assertIn(" line ",
                      test_walker.runtime_errors[0])
        self.assertIn(" col ",
                      test_walker.runtime_errors[0])

    def test_root_type_nodes(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.root_type_nodes)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, ["root", "root"])

    def test_invalid_key_error(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.invalid_key_error)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        errors = test_walker.runtime_errors
        self.assertGreater(len(errors), 0)

    def test_auto_cast(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.auto_cast)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, [True, True])

    def test_no_error_on_dict_key_assign(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.no_error_on_dict_key_assign)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, [{'b': 4}])
        self.assertEqual(len(test_walker.runtime_errors), 0)

    def test_report_status(self):
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.report_status)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, ['hello'])
        self.assertEqual(test_walker.report_status, 302)

    def test_graph_in_graph(self):
        self.logger_on()
        gph = graph(m_id='anon', h=mem_hook())
        sent = sentinel(m_id='anon', h=gph._h)
        sent.register_code(jtc.graph_in_graph)
        test_walker = \
            sent.walker_ids.get_obj_by_name('init')
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(len(report), 3)



"""
Item class for Jaseci

Each item has an id, name, timestamp.
"""
from jaseci.element.element import element


class item(element):
    """Item class for Jaseci"""

    def __init__(self, value=None, *args, **kwargs):
        self.item_value = value
        super().__init__(*args, **kwargs)

    @property
    def value(self):
        return self.item_value

    @value.setter
    def value(self, val):
        self.item_value = val
        self.save()

    def __str__(self):
        if(self.value):
            return super().__str__()+f':{self.value}'
        else:
            return super().__str__()+f':None'



"""
Variable manager for Jac

Representations for all jac runtime variables
"""
from jaseci.element.element import element
from jaseci.graph.node import node
from jaseci.graph.edge import edge
from jaseci.jac.jac_set import jac_set
import uuid

NoneType = type(None)


class JAC_TYPE:
    STR = 'JAC_TYPE.STR'
    INT = 'JAC_TYPE.INT'
    FLOAT = 'JAC_TYPE.FLOAT'
    LIST = 'JAC_TYPE.LIST'
    DICT = 'JAC_TYPE.DICT'
    BOOL = 'JAC_TYPE.BOOL'
    NODE = 'JAC_TYPE.NODE'
    EDGE = 'JAC_TYPE.EDGE'
    TYPE = 'JAC_TYPE.TYPE'
    NULL = 'JAC_TYPE.NULL'


def jac_type_wrap(val):
    if(isinstance(val, type)):
        if(val == str):
            val = JAC_TYPE.STR
        elif(val == int):
            val = JAC_TYPE.INT
        elif(val == float):
            val = JAC_TYPE.FLOAT
        elif(val == list):
            val = JAC_TYPE.LIST
        elif(val == jac_set):
            val = JAC_TYPE.LIST
        elif(val == dict):
            val = JAC_TYPE.DICT
        elif(val == bool):
            val = JAC_TYPE.BOOL
        elif(val == node):
            val = JAC_TYPE.NODE
        elif(val == edge):
            val = JAC_TYPE.EDGE
        elif(val == type):
            val = JAC_TYPE.TYPE
        elif(val == NoneType):
            val = JAC_TYPE.NULL
    return val


def jac_type_unwrap(val):
    if(type(val) == str and val.startswith('JAC_TYPE.')):
        if(val == JAC_TYPE.STR):
            val = str
        elif(val == JAC_TYPE.INT):
            val = int
        elif(val == JAC_TYPE.FLOAT):
            val = float
        elif(val == JAC_TYPE.LIST):
            val = list
        elif(val == JAC_TYPE.DICT):
            val = dict
        elif(val == JAC_TYPE.BOOL):
            val = bool
        elif(val == JAC_TYPE.NODE):
            val = node
        elif(val == JAC_TYPE.EDGE):
            val = edge
        elif(val == JAC_TYPE.TYPE):
            val = type
        elif(val == JAC_TYPE.NULL):
            val = NoneType
    return val


def is_jac_elem(val):
    """Test if value is jac element"""
    return type(val) == str and val.startswith("jac:uuid:")


def jac_elem_wrap(val, serialize_mode=False):
    if(serialize_mode):
        val = val.serialize()
    else:
        val = val.id.urn.replace('urn', 'jac')
    return val


def jac_elem_unwrap(val, parent):
    val = parent._h.get_obj(
        parent._m_id, uuid.UUID(val.replace('jac', 'urn')))
    return val


def jac_wrap_value(val, serialize_mode=False):
    """converts all elements to uuids in lists etc"""
    val = jac_type_wrap(val)
    if (isinstance(val, element)):
        val = jac_elem_wrap(val, serialize_mode=serialize_mode)
    elif(isinstance(val, jac_set)):
        val = jac_wrap_value(list(val), serialize_mode)
    elif (isinstance(val, list)):
        for i in range(len(val)):
            val[i] = jac_wrap_value(val[i], serialize_mode)
    elif (isinstance(val, dict)):
        for i in val.keys():
            val[i] = jac_wrap_value(val[i], serialize_mode)
    return val


def jac_unwrap_value(val, parent):
    """Reference to variables value"""
    if(is_jac_elem(val)):
        val = jac_elem_unwrap(val, parent=parent)
    elif (isinstance(val, list)):
        for i in range(len(val)):
            val[i] = jac_unwrap_value(val[i], parent=parent)
    elif (isinstance(val, dict)):
        for i in val.keys():
            val[i] = jac_unwrap_value(val[i], parent=parent)
    return jac_type_unwrap(val)


class jac_value():
    """
    A reference to a variable in context dict that is common for elements
    """

    def __init__(self, parent, value=None, ctx=None, name=None, end=None):
        """
        Abstraction of all Jac types, ctx and name serve as obj[key/idx]
        end is for idx ranges for list slices
        """
        self.parent = parent
        self.ctx = ctx
        self.is_element = False
        self.name = name
        self.end = end
        self.value = self.setup_value(value)

    def setup_value(self, value):
        if (isinstance(self.ctx, element)):
            self.is_element = self.ctx
            self.ctx = self.ctx.context
        if value is not None:
            return value
        elif(self.ctx is not None and self.name is not None):
            if(self.end is not None):
                return self.ctx[self.name:self.end]
            elif(type(self.name) == int or self.name in self.ctx.keys()):
                return self.ctx[self.name]
        else:
            return None

    def write(self, jac_ast, force=False):
        if(not force and self.is_element and self.name not in self.ctx.keys()
           and not self.parent.parent().
           check_in_arch_context(self.name, self.is_element)):
            self.parent.rt_error(
                f"Creating variable {self.name} in graph "
                f"element {type(self.is_element)} is not allowed, "
                "please define",
                jac_ast)
        elif(self.ctx is None or self.name is None):
            self.parent.rt_error(
                f"No valid live variable! ctx: {self.ctx} name: {self.name}",
                jac_ast)
        elif(self.end is not None):
            self.ctx[self.name:self.end] = self.wrap()
        else:
            self.ctx[self.name] = self.wrap()

    def self_destruct(self, jac_ast):
        if(self.is_element and self.name in self.ctx.keys()):
            self.parent.rt_error(
                f"Deleting {self.name} in graph element "
                f"{type(self.is_element)} is not allowed, try setting to null",
                jac_ast)
            return
        if(self.ctx is not None):
            try:
                del self.ctx[self.name]
            except Exception as e:
                self.parent.rt_error(f'{e}', jac_ast)
        else:
            self.parent.rt_error(
                f'{self.value} is not destroyable',
                jac_ast)

    def wrap(self, serialize_mode=False):
        "Caller for recursive wrap"
        self.value = jac_wrap_value(self.value, serialize_mode)
        return self.value

    def unwrap(self):
        "Caller for recursive unwrap"
        self.value = jac_unwrap_value(self.value, self.parent)
        return self.value

    def jac_type(self):
        """Return Jac type of value"""
        return jac_type_wrap(type(self.value))



"""
Graph api functions as a mixin
"""
from jaseci.api.interface import interface
from jaseci.utils.id_list import id_list
from jaseci.graph.graph import graph
from jaseci.graph.node import node
from jaseci.actor.sentinel import sentinel
import uuid


class graph_api():
    """
    Graph APIs
    """

    def __init__(self):
        self.active_gph_id = None
        self.graph_ids = id_list(self)

    @interface.private_api()
    def graph_create(self, set_active: bool = True):
        """
        Create a graph instance and return root node graph object
        """
        gph = graph(m_id=self._m_id, h=self._h)
        self.graph_ids.add_obj(gph)
        if(set_active):
            self.graph_active_set(gph)
        return gph.serialize()

    @interface.private_api()
    def graph_get(self, gph: graph = None,
                  mode: str = 'default', detailed: bool = False):
        """
        Return the content of the graph with mode
        Valid modes: {default, dot, }
        """
        if(mode == 'dot'):
            return gph.graph_dot_str()
        else:
            items = []
            for i in gph.get_all_nodes():
                items.append(i.serialize(detailed=detailed))
            for i in gph.get_all_edges():
                items.append(i.serialize(detailed=detailed))
            return items

    @interface.private_api()
    def graph_list(self, detailed: bool = False):
        """
        Provide complete list of all graph objects (list of root node objects)
        """
        gphs = []
        for i in self.graph_ids.obj_list():
            gphs.append(i.serialize(detailed=detailed))
        return gphs

    @interface.private_api(cli_args=['gph'])
    def graph_active_set(self, gph: graph):
        """
        Sets the default graph master should use
        """
        self.active_gph_id = gph.jid
        self.alias_register('active:graph', gph.jid)
        return [f'Graph {gph.id} set as default']

    @interface.private_api()
    def graph_active_unset(self):
        """
        Unsets the default sentinel master should use
        """
        self.active_gph_id = None
        self.alias_delete('active:graph')
        return ['Default graph unset']

    @interface.private_api()
    def graph_active_get(self, detailed: bool = False):
        """
        Returns the default graph master is using
        """
        if(self.active_gph_id):
            default = self._h.get_obj(
                self._m_id, uuid.UUID(self.active_gph_id))
            return default.serialize(detailed=detailed)
        else:
            return ['No default graph is selected!']

    @interface.private_api(cli_args=['gph'])
    def graph_delete(self, gph: graph):
        """
        Permanently delete graph with given id
        """
        if(self.active_gph_id == gph.jid):
            self.graph_active_unset()
        self.graph_ids.destroy_obj(gph)
        return [f'Graph {gph.id} successfully deleted']

    @interface.private_api(cli_args=['nd'])
    def graph_node_get(self, nd: node, ctx: list = None):
        """
        Returns value a given node
        """
        ret = {}
        nd_ctx = nd.serialize(detailed=True)['context']
        if(ctx):
            for i in nd_ctx.keys():
                if i in ctx:
                    ret[i] = nd_ctx[i]
        return ret

    @interface.private_api(cli_args=['nd'])
    def graph_node_set(self, nd: node, ctx: dict, snt: sentinel = None):
        """
        Assigns values to member variables of a given node using ctx object
        """
        nd.set_context(
            ctx=ctx, arch=snt.run_architype(
                nd.name, kind='node', caller=self))
        return nd.serialize()

    def active_gph(self):
        return self._h.get_obj(
            self._m_id, uuid.UUID(self.active_gph_id))

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.graph_ids.obj_list():
            i.destroy()



"""
Admin config api functions as a mixin
"""
from jaseci.api.interface import interface


class config_api():
    """
    Admin config APIs
    Abstracted since there are no valid configs in core atm, see jaseci_serv
    to see how used.
    """

    def __init__(self, *args, **kwargs):
        self._valid_configs = ['CONFIG_EXAMPLE', 'STRIPE_KEY', 'ACTION_SETS']

    @interface.admin_api(cli_args=['name'])
    def config_get(self, name: str,
                   do_check: bool = True):
        """
        Get a config
        """
        if(do_check and not self.name_check(name)):
            return self.name_error(name)
        return self._h.get_glob(name)

    @interface.admin_api(cli_args=['name'])
    def config_set(self, name: str, value: str,
                   do_check: bool = True):
        """
        Set a config
        """
        if(do_check and not self.name_check(name)):
            return self.name_error(name)
        self._h.save_glob(name, value)
        return [f"Config of '{name}' to '{value}' set!"]

    @interface.admin_api()
    def config_list(self):
        """
        Check a config is present
        """
        return [v for v in self._h.list_glob() if v in self._valid_configs]

    @interface.admin_api()
    def config_index(self):
        """
        List all valid configs
        """
        return self._valid_configs

    @interface.admin_api(cli_args=['name'])
    def config_exists(self, name: str):
        """
        Check a config is present
        """
        return self._h.has_glob(name)

    @interface.admin_api(cli_args=['name'])
    def config_delete(self, name: str,
                      do_check: bool = True):
        """
        Delete a config
        """
        if(do_check and not self.name_check(name)):
            return self.name_error(name)
        self._h.destroy_glob(name)
        return [f"{name} Deleted."]

    def name_error(self, name):
        """Much used error output"""
        return [
            f"Config {name} not recognized, must be in {self._valid_configs}!"]

    def name_check(self, name):
        """Much used name check"""
        return (name in self._valid_configs)



import sys
import io

from jaseci.utils.mem_hook import mem_hook
from jaseci.actor.sentinel import sentinel
from jaseci.graph.graph import graph
import jaseci.jac.tests.dot_code as dtc

from unittest import TestCase
from jaseci.utils.utils import TestCaseHelper


class test_dot(TestCaseHelper, TestCase):
    """Unit test for DOT support"""

    def setUp(self):
        super().setUp()
        self.gph = graph(m_id='anon', h=mem_hook())
        self.sent = sentinel(m_id=self.gph._m_id, h=self.gph._h)
        self.old_stdout = sys.stdout
        self.new_stdout = io.StringIO()
        sys.stdout = self.new_stdout

    def tearDown(self):
        sys.stdout = self.old_stdout
        super().tearDown()

    def to_screen(self):
        sys.stdout = self.old_stdout
        print("output: ", self.new_stdout.getvalue())
        sys.stdout = self.new_stdout

    def test_dot_node(self):
        """Test node in dot"""
        self.sent.register_code(dtc.dot_node)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "graph_root_node_name\n")

    def test_dot_node_no_dot_id(self):
        """Test node in dot with no do graph name"""
        self.sent.register_code(dtc.dot_node_no_dot_id)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "graph_root_node_name\n")

    def test_spawn_graph_node(self):
        """Test node in using spawn graphs instead of dot"""
        self.sent.register_code(dtc.spawn_graph_node)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "graph_root_node_name\n")

    def test_dot_node_multi_stmts(self):
        """Test node in dot, defined with multiple statements."""
        self.sent.register_code(dtc.dot_node_multi_stmts)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(),
                         "real_test_node\n"
                         "2021\n")

    def test_dot_edge(self):
        """Test edge in dot."""
        self.sent.register_code(dtc.dot_edge)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(),
                         "root\n"
                         "node_1\n"
                         "node_2\n")

    def test_dot_edge_with_attrs(self):
        """Test edge in dot with attrs"""
        self.sent.register_code(dtc.dot_edge_with_attrs)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(),
                         "root\n"
                         "node_1\n")

    def test_dot_edge_with_attrs_vars(self):
        """Test edge in dot with attrs and variables"""
        self.sent.register_code(dtc.dot_edge_with_attrs_vars)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(),
                         "edge_1\n"
                         "node_1\n")

    def test_dot_graph_parses(self):
        self.sent.register_code(dtc.dot_graph)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()

    def test_dot_str(self):
        self.sent.register_code(dtc.dot_graph)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        dot_str = self.gph.graph_dot_str()
        self.assertTrue('strict digraph root' in dot_str)

    def test_dot_quoted_string(self):
        self.sent.register_code(dtc.dot_quoted_string)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        report = gen_walker.report
        self.assertEqual(
            report, ['root', 'this has space', 'another space'])



"""
Node class for Jaseci

Each node has an id, name, timestamp and it's set of edges.
First node in list of 'member_node_ids' is designated root node
"""
from jaseci.element.element import element
from jaseci.element.obj_mixins import anchored
from jaseci.graph.edge import edge
from jaseci.utils.id_list import id_list
from jaseci.utils.utils import logger

import uuid


class node(element, anchored):
    """Node class for Jaseci"""

    def __init__(self, dimension=0, *args, **kwargs):
        self.edge_ids = id_list(self)
        self.parent_node_ids = id_list(self)
        self.member_node_ids = id_list(self)
        self.dimension = dimension  # Nodes are always hdgd 0
        self.context = {}

        anchored.__init__(self)
        element.__init__(self, *args, **kwargs)

    def attach(self, node_obj, edge_set=None, as_outbound=True,
               as_bidirected=False):
        """
        Generalized attach function for attaching nodes with edges
        """
        if(edge_set is None):
            edge_set = [edge(m_id=self._m_id, h=self._h,
                             kind='edge', name='generic')]
        link_order = [self, node_obj] if as_outbound else [node_obj, self]
        for e in edge_set:
            if(not e.set_from_node(link_order[0]) or
               not e.set_to_node(link_order[1])):
                # Node not found error logged in set node function
                return []
            e.set_bidirected(as_bidirected)
        # save and return
        self.save()
        node_obj.save()
        return edge_set

    def attach_outbound(self, node_obj, edge_set=None):
        """
        Creates edge to a node and returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_outbound=True)

    def attach_inbound(self, node_obj, edge_set=None):
        """
        Creates edge from a node and returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_outbound=False)

    def attach_bidirected(self, node_obj, edge_set=None):
        """
        Creates bidirected edge to node returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_bidirected=True)

    def detach(self, node_obj, edge_set=None, as_outbound=True,
               as_bidirected=False, ignore_direction=False, silent=True):
        """
        Generalized detach function for detaching nodes with edges
        """
        if(edge_set is None):
            edge_set = self.attached_edges(node_obj)
        link_order = [self, node_obj] if as_outbound else [node_obj, self]
        num_detached = 0
        for e in edge_set:
            # validate edge connection exists
            if(not e.connects(link_order[0], link_order[1],
                              ignore_direction=ignore_direction)):
                if(not silent):
                    logger.warning(
                        str(
                            f"{e} does not connect "
                            f"{link_order[0]} to {link_order[1]}")
                    )
                continue
            if(as_bidirected and not ignore_direction
               and not e.is_bidirected()):
                if(not silent):
                    logger.warning(
                        str(f"{e} is not a bidirected edge "))
                continue
            # destroy edge
            num_detached += 1
            e.destroy()
        return num_detached

    def detach_outbound(self, node_obj, edge_set=None, silent=True):
        """
        Destroy edges to a node
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_outbound=True)

    def detach_inbound(self, node_obj, edge_set=None, silent=True):
        """
        Destroy edges from a node
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_outbound=False)

    def detach_bidirected(self, node_obj, edge_set=None, silent=True):
        """
        Destroy bidirected edges between nodes
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_bidirected=True)

    def detach_edges(self, node_obj, edge_set=None, silent=True):
        """
        Destroy given edges between nodes without checking orientation
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, ignore_direction=True)

    def destroy_outbound(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if(edge_set is None):
            edge_set = self.outbound_edges(node_obj)
        if(self.detach_outbound(node_obj, edge_set)):
            node_obj.destroy()

    def destroy_inbound(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if(edge_set is None):
            edge_set = self.inbound_edges(node_obj)
        if(self.detach_inbound(node_obj, edge_set)):
            node_obj.destroy()

    def destroy_bidirected(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if(edge_set is None):
            edge_set = self.bidirected_edges(node_obj)
        if(self.detach_bidirected(node_obj, edge_set)):
            node_obj.destroy()

    def is_attached_out(self, node_obj, edge_set=None):
        """
        Tests whether edges attach to a node
        """
        out_set = self.outbound_edges(node_obj)
        if(edge_set is None):
            return len(out_set)
        else:
            for e in edge_set:
                if(e not in out_set):
                    return False
        return True

    def is_attached_in(self, node_obj, edge_set=None):
        """
        Tests whether edges attach from a node
        """
        in_set = self.inbound_edges(node_obj)
        if(edge_set is None):
            return len(in_set)
        else:
            for e in edge_set:
                if(e not in in_set):
                    return False
        return True

    def is_attached_bi(self, node_obj, edge_set=None):
        """
        Tests whether edges attach either to or from a node
        """
        bi_set = self.bidirected_edges(node_obj)
        if(edge_set is None):
            return len(bi_set)
        else:
            for e in edge_set:
                if(e not in bi_set):
                    return False
        return True

    def outbound_edges(self, node_obj=None):
        """Returns list of all edges out of node"""
        edge_set = []
        for e in self.edge_ids.obj_list():
            if(not e.is_bidirected() and e.connects(self, node_obj)):
                edge_set.append(e)
        return edge_set

    def inbound_edges(self, node_obj=None):
        """Returns list of all edges in to node"""
        edge_set = []
        for e in self.edge_ids.obj_list():
            if(not e.is_bidirected() and e.connects(node_obj, self)):
                edge_set.append(e)
        return edge_set

    def bidirected_edges(self, node_obj=None):
        """Returns list of all edges between nodes"""
        edge_set = []
        for e in self.edge_ids.obj_list():
            if(e.is_bidirected() and e.connects(self, node_obj)):
                edge_set.append(e)
        return edge_set

    def attached_edges(self, node_obj=None, silent=False):
        """
        Returns the edges connecting self to or from a node

        silent is used to indicate whther the edge is intened to be used.
        (effectively turns off error checking when false)
        """
        edge_set = self.outbound_edges(node_obj) + \
            self.inbound_edges(node_obj) + \
            self.bidirected_edges(node_obj)
        if(not silent and edge_set is None):
            logger.error(
                str(f"No edges found between {self} and {node_obj}")
            )
        return edge_set

    def outbound_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges out"""
        if(edge_set is None):
            edge_set = self.edge_ids.obj_list()
        ret_list = []
        for e in edge_set:
            if (not e.is_bidirected() and e.connects(source=self)):
                ret_list.append(e.to_node())
        return ret_list

    def inbound_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges in"""
        if(edge_set is None):
            edge_set = self.edge_ids.obj_list()
        ret_list = []
        for e in edge_set:
            if (not e.is_bidirected() and e.connects(target=self)):
                ret_list.append(e.from_node())
        return ret_list

    def bidirected_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges"""
        if(edge_set is None):
            edge_set = self.edge_ids.obj_list()
        ret_list = []
        for e in edge_set:
            if (e.is_bidirected()):
                ret_list.append(e.opposing_node(self))
        return ret_list

    def attached_nodes(self):
        """Returns list of all nodes connected"""
        edge_set = self.edge_ids.obj_list()
        ret_list = []
        for e in edge_set:
            ret_list.append(e.opposing_node(self))
        return ret_list

    def dimension_matches(self, node_obj, silent=True):
        """Test if dimension matches another node"""
        matches = self.dimension == node_obj.dimension
        if(not matches and not silent):
            logger.error(
                str("'{}' cant connect to '{}' - dim mismatch {}->{}".
                    format(self, node_obj, self.dimension,
                           node_obj.dimension))
            )
        return matches

    def make_member_of(self, node_obj):
        """
        Adds node to higher dimension node and does relevant checks
        """
        # check if valid inclusion
        if(node_obj.dimension != self.dimension+1):
            logger.error(
                str("'{}' cant be member to '{}' - dimension mismatch {}->{}".
                    format(self, node_obj, self.dimension, node_obj.dimension))
            )
        # adds self to hdgd and hdgd to list of owners
        else:
            node_obj.member_node_ids.add_obj(self)
            self.parent_node_ids.add_obj(node_obj)

    def make_owner_of(self, node_obj):
        """
        Adds node to lower dimension node and does relevant checks
        """
        node_obj.make_member_of(self)

    def leave_memebership_of(self, node_obj):
        """Remove node from higher dimension node"""
        node_obj.member_node_ids.remove_obj(self)
        self.parent_node_ids.remove_obj(node_obj)

    def disown(self, node_obj):
        """Remove node from higher dimension node"""
        node_obj.leave_membership_of(self)

    def set_context(self, ctx, arch=None):
        """Assign values to context fields of node, arch is node architype"""
        if(arch is None):
            arch = self
        for i in ctx.keys():
            if (i not in arch.context.keys()):
                logger.warning(str(f"{i} not a context member of {self}"))
                continue
            else:
                self.context[i] = ctx[i]
        self.save()

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.edge_ids.obj_list():
            i.destroy()
        super().destroy()

    def dot_str(self, node_map=None):
        """
        DOT representation
        """
        if(node_map is None):
            nid = f'{uuid.UUID(self.jid).hex}'
        else:
            nid = f'{node_map.index(self.jid)}'

        dstr = f'"n{nid}" [ id="{uuid.UUID(self.jid).hex}", '
        dstr += f'label="n{nid}:{self.name}" '

        node_dict = self.context

        if (node_dict):
            for k, v in node_dict.items():
                if(not isinstance(v, str) or v == ""):
                    continue
                dstr += f', {k}="{v[:32]}"'

        dstr += ' ]'

        return dstr+'\n'



from enum import Enum, auto


class ref(Enum):
    SELF = auto()
    ENTRY_ACT_IDS = auto()
    EXIT_ACT_IDS = auto()
    ACTIV_ACT_IDS = auto()
    STOPPED = auto()
    RESULT_OUT = auto()
    TMP1 = auto()
    TMP2 = auto()
    TMP3 = auto()

    # Walker refs
    HERE_ID = auto()
    STEP = auto()
    IN_ENT_EXIT = auto()
    NEXT_N_IDS = auto()
    IGNORE_N_IDS = auto()
    DESTROY_N_IDS = auto()

    # Arhcitype refs



"""
Graph  class for Jaseci

"""
from jaseci.graph.node import node
from jaseci.utils.id_list import id_list


class graph(node):
    """Graph class for Jaseci"""

    def __init__(self, *args, **kwargs):
        self.hd_node_ids = id_list(self)
        node.__init__(self, name='root', kind='node', *args, **kwargs)

    def get_all_nodes(self, node_list=None):
        """
        Returns all reachable nodes
        node_list is used internally for recursion
        """
        if(not isinstance(node_list, list)):
            node_list = []

        # if cycle detected in path
        if(self in node_list):
            return node_list

        node_list.append(self)

        for i in self.attached_nodes():
            graph.get_all_nodes(i, node_list)

        return node_list

    def get_all_edges(self):
        """
        Returns all reachable edges
        """
        edge_set = set()
        node_list = self.get_all_nodes()

        for i in node_list:
            for e in i.attached_edges():
                edge_set.add(e)

        return list(edge_set)

    def graph_dot_str(self):
        """
        DOT representation for graph.
        NOTE: This is different from the dot_str method for node intentionally
        because graph inherits node.
        """
        node_list = self.get_all_nodes()
        edge_list = self.get_all_edges()
        node_map = [i.jid for i in node_list]
        edge_map = [i.jid for i in edge_list]

        # Construct the graph string
        dstr = ''
        dstr += f'strict digraph {self.name} {{\n'
        for n in node_list:
            dstr += f'    {n.dot_str(node_map)}'
        for e in edge_list:
            dstr += f'    {e.dot_str(node_map, edge_map)}'
        dstr += '}'
        return dstr

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.hd_node_ids.obj_list():
            i.destroy()
        super().destroy()



from unittest import TestCase
from jaseci.utils.utils import TestCaseHelper
import jaseci.actions.live_actions as jla
import jaseci.actions.remote_actions as jra


class jac_actions_tests(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_remote_action_example(self):
        from typing import Union
        @jla.jaseci_action(act_group=['use'], aliases=['enc_question'],
                           allow_remote=True)
        def question_encode(question: Union[str, list]):
            pass

        @jla.jaseci_action(act_group=['use'], aliases=['enc_answer'],
                           allow_remote=True)
        def answer_encode(answer: Union[str, list],
                          context: Union[str, list] = None):
            pass

        @jla.jaseci_action(act_group=['use'],
                           allow_remote=True)
        def cos_sim_score(q_emb: list, a_emb: list):
            pass

        @jla.jaseci_action(act_group=['use'], aliases=['qa_score'],
                           allow_remote=True)
        def dist_score(q_emb: list, a_emb: list):
            pass
        jra.serv_actions()
        self.assertEqual(jra.remote_actions,
                         {'use.answer_encode': ['answer', 'context'],
                          'use.cos_sim_score': ['q_emb', 'a_emb'],
                          'use.dist_score': ['q_emb', 'a_emb'],
                          'use.enc_answer': ['answer', 'context'],
                          'use.enc_question': ['question', ],
                          'use.qa_score': ['q_emb', 'a_emb'],
                          'use.question_encode': ['question']})

    def test_live_action_globals(self):
        self.assertGreater(len(jla.live_actions), 25)



from .mem_hook import mem_hook
import pickle


class file_hook(mem_hook):
    """
    Uses files for persistence
    """

    def __init__(self, filename):
        mem_hook.__init__(self)
        self.filename = filename
        with open(filename, 'rb') as f:
            self.mem = pickle.load(f)

    def commit(self):
        """Write through all saves to store"""
        with open(self.filename, 'wb') as f:
            pickle.dump(self.mem, f)



"""Built in actions for Jaseci"""
from datetime import datetime
from datetime import timedelta
from jaseci.jac.machine.jac_value import jac_type_unwrap as jtu
from jaseci.actions.live_actions import jaseci_action


@jaseci_action()
def quantize_to_year(date: str):
    date = datetime.fromisoformat(jtu(date))
    date = date.replace(month=1, day=1, hour=0, minute=0,
                        second=0, microsecond=0)
    return date.isoformat()


@jaseci_action()
def quantize_to_month(date: str):
    date = datetime.fromisoformat(jtu(date))
    date = date.replace(day=1, hour=0, minute=0,
                        second=0, microsecond=0)
    return date.isoformat()


@jaseci_action()
def quantize_to_week(date: str):
    date = datetime.fromisoformat(jtu(date))
    date = date.replace(hour=0, minute=0,
                        second=0, microsecond=0)
    date = date - timedelta(days=date.weekday())
    return date.isoformat()


@jaseci_action()
def quantize_to_day(date: str):
    date = datetime.fromisoformat(jtu(date))
    date = date.replace(hour=0, minute=0,
                        second=0, microsecond=0)
    return date.isoformat()


@jaseci_action()
def date_day_diff(start_date: str, end_date: str):
    # Try to deal with some incompatible date in old nodes
    # Due date has not been saving in isoformat so this doesn't work
    # for example,'2021-03-31T04:00:00.000Z'
    start_date = datetime.fromisoformat(jtu(start_date.split('T')[0]))
    end_date = datetime.fromisoformat(jtu(end_date.split('T')[0]))

    delta = end_date - start_date

    return delta.days



"""Built in actions for Jaseci"""
import requests
from jaseci.actions.live_actions import jaseci_action


@jaseci_action()
def get(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.get(url, json=data, headers=header)
    ret = {'status_code': res.status_code}
    try:
        ret['response'] = res.json()
    except Exception:
        ret['response'] = res.text
    return ret


@jaseci_action()
def post(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.post(url, json=data, headers=header)
    ret = {'status_code': res.status_code}
    try:
        ret['response'] = res.json()
    except Exception:
        ret['response'] = res.text
    return ret


@jaseci_action()
def put(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.put(url, json=data, headers=header)
    ret = {'status_code': res.status_code}
    try:
        ret['response'] = res.json()
    except Exception:
        ret['response'] = res.text
    return ret


@jaseci_action()
def delete(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.delete(url, json=data, headers=header)
    ret = {'status_code': res.status_code}
    try:
        ret['response'] = res.json()
    except Exception:
        ret['response'] = res.text
    return ret


@jaseci_action()
def head(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.head(url, json=data, headers=header)
    ret = {'status_code': res.status_code}
    try:
        ret['response'] = res.json()
    except Exception:
        ret['response'] = res.text
    return ret


@jaseci_action()
def options(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.options(url, json=data, headers=header)
    ret = {'status_code': res.status_code}
    try:
        ret['response'] = res.json()
    except Exception:
        ret['response'] = res.text
    return ret



"""
Mix in for jac code object in Jaseci
"""
import json
from jaseci.utils.utils import logger
from jaseci.jac.ir.ast import ast
import hashlib


class jac_json_enc(json.JSONEncoder):
    """Custom Json encoder for Jac ASTs"""

    def default(self, obj):
        if(isinstance(obj, ast)):
            retd = {}
            for i in obj.__dict__.keys():
                if(not i.startswith('_')):
                    retd[i] = obj.__dict__[i]
            return retd
        return super().default(obj)


class jac_json_dec(json.JSONDecoder):
    """Custom hook for decoding Jac ASTs"""

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(
            self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):

        if isinstance(obj, dict) and "mod_name" in obj and "kid" in obj:
            ret = ast(mod_name=obj['mod_name'], fresh_start=False)
            for i in obj.keys():
                setattr(ret, i, obj[i])
            return ret
        return obj


def jac_ast_to_ir(jac_ast):
    """Convert AST to IR string"""
    return json.dumps(cls=jac_json_enc, obj=jac_ast)


def jac_ir_to_ast(ir):
    """Convert AST to IR string"""
    return json.loads(cls=jac_json_dec, s=ir)


class jac_code():
    """Obj mixin to code pickling"""

    def __init__(self, code_ir=None):
        self.is_active = False
        self.code_ir = None
        self.code_sig = None
        self._jac_ast = None
        self.errors = []
        self.apply_ir(code_ir)

    def reset(self):
        self.is_active = False

    def refresh(self):
        self._jac_ast = jac_ir_to_ast(self.code_ir) if self.code_ir else None
        if(self._jac_ast):
            self.is_active = True
        else:
            self.is_active = False

    def apply_ir(self, ir):
        """Apply's IR to object"""
        self.code_ir = ir if(isinstance(ir, str)) else \
            jac_ast_to_ir(ir)
        self.code_sig = hashlib.md5(self.code_ir.encode()).hexdigest()
        jac_code.refresh(self)  # should disregard overloaded versions
        if(self._jac_ast and
           (self.j_type == 'architype' or self.j_type == 'walker') and
           (self._jac_ast.name == 'architype' or
                self._jac_ast.name == 'walker')):
            kid = self._jac_ast.kid
            self.kind = f"{kid[0].token_text()}"
            self.name = f"{kid[1].token_text()}"

    def parse_jac(self, code, dir, start_rule='start'):
        """Generate AST tree from Jac code text"""
        tree = ast(jac_text=code, start_rule=start_rule, mod_name=self.name,
                   mod_dir=dir)
        self.errors = tree._parse_errors
        if(tree._parse_errors):
            logger.error(str(f'{self.name}: Invalid syntax in Jac code!'))
            for i in tree._parse_errors:
                logger.error(i)
            return None
        return tree

    def register(self, code, dir):
        """
        Parses Jac code and saves IR
        """
        start_rule = 'start' if self.j_type == 'sentinel' else self.j_type
        tree = self.parse_jac(code, dir, start_rule=start_rule)

        if(not tree):
            self.is_active = False
        else:
            self.apply_ir(tree)

        if(not self.is_active):
            logger.error(str(f'{self.name}: Code not registered'))
        return self.is_active

    def ir_dict(self):
        """Return IR as dictionary"""
        return json.loads(self.code_ir)



"""
Jac tools api functions as a mixin
"""
from jaseci.api.interface import interface
import os
import json


class jac_api():
    """
    Jac tool APIs
    """
    @interface.cli_api(cli_args=['file'])
    def jac_build(self, file: str, out: str = ""):
        """
        Command line tooling for building executable jac ir
        """
        if(not os.path.isfile(file)):
            ret = "File does not exsist!"
            return ret
        filename = os.path.basename(file)
        dir = os.path.dirname(os.path.realpath(file))
        if(not len(out)):
            if(file.endswith(".jac")):
                out = file.replace(".jac", ".jir")
            else:
                out = file+'.jir'
        faux = self.faux_master()
        with open(file, 'r') as file:
            ret = faux.sentinel_register(
                code=file.read(), code_dir=dir, name=filename)
            if('success' in ret and not ret['success']):
                return ret
            with open(out, 'w') as ofile:
                jir_out = json.dumps(
                    faux.sentinel_get(mode='ir',
                                      snt=faux.active_snt()))
                ofile.write(jir_out)
                ret = f"Build of {out} complete!"

                return ret

    @interface.cli_api(cli_args=['file'])
    def jac_test(self, file: str, detailed: bool = False):
        """
        Command line tooling for running all test in both .jac code files
        and .jir executables
        """
        if(not os.path.isfile(file)):
            ret = "File does not exsist!"
            return ret
        filename = os.path.basename(file)
        dir = os.path.dirname(os.path.realpath(file))
        is_jir = file.endswith(".jir")
        faux = self.faux_master()
        with open(file, 'r') as file:
            if(is_jir):
                faux.sentinel_register(name=filename)
                faux.sentinel_set(snt=faux.active_snt(),
                                  code=file.read(), mode='ir')
            else:
                ret = faux.sentinel_register(
                    code=file.read(), code_dir=dir,  name=filename,
                    auto_run='')
                if('success' in ret and not ret['success']):
                    return ret
        return faux.sentinel_test(snt=faux.active_snt(), detailed=detailed)

    @interface.cli_api(cli_args=['file'])
    def jac_run(self, file: str, walk: str = 'init', ctx: dict = {},
                profiling: bool = False):
        """
        Command line tooling for running all test in both .jac code files
        and .jir executables
        """
        if(not os.path.isfile(file)):
            ret = "File does not exsist!"
            return ret
        filename = os.path.basename(file)
        dir = os.path.dirname(os.path.realpath(file))
        is_jir = file.endswith(".jir")
        faux = self.faux_master()
        with open(file, 'r') as file:
            if(is_jir):
                faux.sentinel_register(name=filename)
                faux.sentinel_set(snt=faux.active_snt(),
                                  code=file.read(), mode='ir')
            else:
                ret = faux.sentinel_register(
                    code=file.read(), code_dir=dir, name=filename, auto_run='')
                if('success' in ret and not ret['success']):
                    return ret
        return faux.walker_run(name=walk, snt=faux.active_snt(),
                               nd=faux.active_gph(), ctx=ctx,
                               profiling=profiling)

    @interface.cli_api(cli_args=['file'])
    def jac_dot(self, file: str, walk: str = 'init', ctx: dict = {},
                profiling: bool = False):
        """
        Command line tooling for a walker then output graph in both .jac code
        files and .jir executables
        """
        if(not os.path.isfile(file)):
            ret = "File does not exsist!"
            return ret
        filename = os.path.basename(file)
        dir = os.path.dirname(os.path.realpath(file))
        is_jir = file.endswith(".jir")
        faux = self.faux_master()
        with open(file, 'r') as file:
            if(is_jir):
                faux.sentinel_register(name=filename)
                faux.sentinel_set(snt=faux.active_snt(),
                                  code=file.read(), mode='ir')
            else:
                ret = faux.sentinel_register(
                    code=file.read(), name=filename, code_dir=dir,
                    auto_run='')
                if('success' in ret and not ret['success']):
                    return ret
        faux.walker_run(name=walk, snt=faux.active_snt(),
                        nd=faux.active_gph(), ctx=ctx,
                        profiling=profiling)
        return faux.graph_get(gph=faux.active_gph(), mode='dot')

    def faux_master(self):
        from jaseci.element.super_master import super_master
        from jaseci.utils.mem_hook import mem_hook
        return super_master(h=mem_hook())



"""
Base Jaseci object class

Naming conventions
* '_ids', '_id' - appended to lists of ids as feilds of jaseci objects
* 'obj' - used to represent a jaseci object
* 'item' - Items are kv pairs derived from item class, dont use elsewhere
* 'store' - refers to persistent store for hooks from engine, (dont use db)
"""

import uuid
from datetime import datetime
import copy
import json
from jaseci.utils.id_list import id_list
from jaseci.utils.utils import logger, log_var_out
from jaseci.utils.mem_hook import mem_hook
from jaseci.utils.mem_hook import json_str_to_jsci_dict
from jaseci.element.obj_mixins import hookable


__version__ = '1.0.0'


class element(hookable):
    """
    Base class for Jaseci for standard info shared across all objects types in
    Jaseci. This class also includes a method for dumping the non-general items
    as a single dictionary. The :meth:`jsci_payload` function automatically
    finds relevant non general fields and returns a dictionary of them.

    Keep in mind that when a valid hook is passed, the item writes itself on
    creation.

    Auto save should be faulse when loading from database so a new UUID is
    not randomly created and saved back to db on load

    TODO: Make below content a documentation variable and link in a few places
    Keep in mind the postfix of '_ids' on feilds of derived classes of element
    is a Jaseci convention for lists of uuids that are stored as hex and loaded
    as type UUID.

    NOTE: All fields of elements and it's derived classes are assumed to be
    json serializable types
    """

    def __init__(self, m_id, h, parent_id=None, name='basic',
                 kind='generic', auto_save=True, *args, **kwargs):
        self.name = name
        self.kind = kind
        self.jid = uuid.uuid4().urn
        self.j_parent = parent_id.urn if parent_id else None  # member of
        self.j_timestamp = datetime.utcnow().isoformat()
        self.j_type = type(self).__name__
        if(self.is_master()):
            m_id = self.jid
        hookable.__init__(self, h, m_id,  *args, **kwargs)
        if(auto_save):
            self.save()

    @property
    def id(self):
        return uuid.UUID(self.jid)

    @id.setter
    def id(self, obj):
        self.jid = obj.urn

    @property
    def parent_id(self) -> uuid.UUID:
        if (not self.j_parent):
            return None
        return uuid.UUID(self.j_parent)

    @parent_id.setter
    def parent_id(self, obj: uuid.UUID):
        if (not obj):
            self.j_parent = None
        else:
            self.j_parent = obj.urn

    @property
    def timestamp(self):
        return datetime.fromisoformat(self.j_timestamp)

    @timestamp.setter
    def timestamp(self, obj):
        self.j_timestamp = obj.isoformat()

    def is_master(self, super_check=False, silent=True):
        """Check if self is a master"""
        ret = True
        if('master' not in self.j_type):
            ret = False
            if(not silent):
                logger.error(f'{self} is not master')
        if(super_check and 'super' not in self.j_type):
            ret = False
            if(not silent):
                logger.error(f'{self} does not have super master status')
        return ret

    def duplicate(self, persist_dup: bool = False):
        """
        Duplicates elements by creating copy with new id
        Hook override to duplicate into mem / another store
        """

        dup = type(self)(m_id=self._m_id, h=self._h, persist=persist_dup)
        id_save = dup.id
        for i in dup.__dict__.keys():
            if(type(dup.__dict__[i]) == id_list):
                setattr(dup, i, id_list(
                    parent_obj=dup, in_list=self.__dict__[i]))
            else:
                setattr(dup, i, self.__dict__[i])
        dup.id = id_save
        dup.timestamp = datetime.utcnow()
        dup.save()
        return dup

    def is_equivalent(self, obj):
        """
        Duplicates elements by creating copy with new id
        """
        if(self.j_type != obj.j_type):
            return False
        for i in vars(self).keys():
            if (not i.startswith('_') and not callable(getattr(self, i))):
                if(i != 'jid' and i != 'j_timestamp'):
                    if (getattr(self, i) != getattr(obj, i)):
                        return False
        return True

    def jsci_payload(self):
        """
        Returns all data fields and values of jaseci object as json string.
        This grabs any fields that are added into inherited objects. Useful for
        saving and loading item.
        """
        obj_fields = []
        element_fields = dir(element(m_id=self._m_id, h=mem_hook()))
        for i in vars(self).keys():
            if not i.startswith('_') and i not in element_fields:
                obj_fields.append(i)
        obj_dict = {}
        for i in obj_fields:
            obj_dict[i] = getattr(self, i)
        return json.dumps(obj_dict)

    def serialize(self, deep=0, detailed=False):
        """
        Serialize Jaseci object
        """
        jdict = {}
        key_fields = ['name', 'kind', 'jid', 'j_type', 'context',
                      'anchor', 'j_timestamp', 'version']
        for i in vars(self).keys():
            if not i.startswith('_'):
                if(not detailed and i not in key_fields):
                    continue
                jdict[i] = copy.copy(vars(self)[i])
                if(not detailed and i == 'context'):
                    if('_private' in jdict[i].keys()):
                        for j in jdict[i]['_private']:
                            del jdict[i][j]
                if (deep > 0 and isinstance(jdict[i], id_list)):
                    for j in range(len(jdict[i])):
                        jdict[i][j] = copy.copy(
                            self._h.get_obj(self._m_id, uuid.UUID(
                                jdict[i][j])).serialize(deep - 1))
        return jdict

    def json(self, deep=0, detailed=False):
        """
        Returns entire self object as Json string

        deep indicates number of levels to unwind uuids
        """
        return json.dumps(self.serialize(deep, detailed=detailed),
                          indent=4)

    def json_load(self, blob):
        """Loads self from json blob"""
        jdict = json_str_to_jsci_dict(blob, parent_obj=self)
        self.dict_load(jdict=jdict)

    def dict_load(self, jdict):
        """Loads self from dict"""
        for i in jdict.keys():
            setattr(self, i, jdict[i])

    def get_deep_obj_list(self, objs=None):
        """Recursively get all contained Jaseci objects and return id_list"""
        if(objs is None):
            objs = []
        objs.append(self)
        for i in self.__dict__.keys():
            if(str(i).endswith("_ids") and
               isinstance(self.__dict__[i], id_list)):
                for j in self.__dict__[i].obj_list():
                    if(j not in objs):
                        j.get_deep_obj_list(objs=objs)
        return objs

    def log_output(self, val):
        """
        Utility to log out variable in pretty format
        """
        log_var_out(val)

    def __str__(self):
        """
        String representation is of the form type:name:kind
        """
        return self.j_type + ':' + self.kind + ':' + self.name + ':' + self.jid

    def __repr__(self):
        return self.__str__()



"""
Sentinel class for Jaseci

Each sentinel has an id, name, timestamp and it's set of walkers.
"""
from jaseci.element.element import element
from jaseci.utils.utils import logger
from jaseci.utils.id_list import id_list
from jaseci.jac.ir.jac_code import jac_code, jac_ir_to_ast
from jaseci.jac.interpreter.sentinel_interp import sentinel_interp
from jaseci.actor.walker import walker
from jaseci.actor.architype import architype


class sentinel(element, jac_code, sentinel_interp):
    """
    Sentinel class for Jaseci

    is_active is used to signify whether sentinel is ready to run walkers, i.e,
    register_code succeeded
    """

    def __init__(self, *args, **kwargs):
        self.version = None
        self.arch_ids = id_list(self)
        self.walker_ids = id_list(self)
        self.testcases = []
        element.__init__(self, *args, **kwargs)
        jac_code.__init__(self, code_ir=None)
        sentinel_interp.__init__(self)
        self.load_arch_defaults()

    def load_arch_defaults(self):
        self.arch_ids.add_obj(architype(m_id=self._m_id, h=self._h,
                                        name='root', kind='node'))
        self.arch_ids.add_obj(architype(m_id=self._m_id, h=self._h,
                                        name='generic', kind='node'))
        self.arch_ids.add_obj(architype(m_id=self._m_id, h=self._h,
                                        name='generic', kind='edge'))

    def reset(self):
        """Resets state of sentinel and unregister's code"""
        self.arch_ids.destroy_all()
        self.load_arch_defaults()
        self.walker_ids.destroy_all()
        jac_code.reset(self)
        sentinel_interp.reset(self)

    def refresh(self):
        super().refresh()
        self.ir_load()

    def register_code(self, text, dir='./'):
        """
        Registers a program (set of walkers and architypes) written in Jac
        """
        self.reset()
        self.register(text, dir)
        if(self.is_active):
            self.ir_load()
        return self.is_active

    def ir_load(self):
        """
        Load walkers and architypes from IR
        """

        self.run_start(self._jac_ast)

        if(self.runtime_errors):
            logger.error(
                str(f'{self.name}: Runtime problem processing sentinel!')
            )
            self.is_active = False
        elif(not self.walker_ids and not self.arch_ids):
            logger.error(
                str(f'{self.name}: No walkers nor architypes created!')
            )
            self.is_active = False
        return self.is_active

    def register_walker(self, code):
        """Adds a walker based on jac code"""
        tree = self.parse_jac(code, start_rule='walker')
        if(not tree):
            return None
        return self.load_walker(tree)

    def register_architype(self, code):
        """Adds a walker based on jac code"""
        tree = self.parse_jac(code, start_rule='architype')
        if(not tree):
            return None
        return self.load_architype(tree)

    def spawn_walker(self, name, caller=None):
        """
        Spawns a new walker from registered walkers and adds to
        live walkers
        """
        src_walk = self.walker_ids.get_obj_by_name(name)
        if (not src_walk):
            logger.error(
                str(f'{self.name}: Unable to spawn walker {name}!')
            )
            return None
        new_walk = src_walk.duplicate()
        if(caller):
            new_walk.set_master(caller._m_id)
        new_walk._jac_ast = src_walk._jac_ast
        if(new_walk._jac_ast is None):
            new_walk.refresh()
        return new_walk

    def spawn_architype(self, name, kind=None, caller=None):
        """
        Spawns a new architype from registered architypes and adds to
        live walkers
        """
        src_arch = self.arch_ids.get_obj_by_name(name, kind=kind,
                                                 silent=True)
        if (not src_arch):
            return None

        if(caller and caller._m_id != src_arch._m_id):
            new_arch = src_arch.duplicate()
            new_arch.set_master(caller._m_id)
            new_arch._jac_ast = src_arch._jac_ast
            if(new_arch._jac_ast is None):
                new_arch.refresh()
            return new_arch
        else:
            return src_arch

    def run_architype(self, name, kind=None, caller=None):
        """
        Spawn, run, then destroy architype if m_id's are different
        """
        if(caller is None):
            caller = self
        arch = self.spawn_architype(name, kind, caller)
        if(arch is None):
            logger.error(
                str(
                    f'{self.name}: Unable to spawn architype '
                    f'{[name, kind]}!')
            )
            return None
        if(arch.jid in self.arch_ids):
            return arch.run()
        else:
            ret = arch.run()
            arch.destroy()
            return ret

    def check_in_arch_context(self, key_name, object):
        """
        Validates a key is present in currently loaded architype of
        a particular instance (helper for expanding node has variables)
        """
        src_arch = self.run_architype(object.name, kind=object.kind)
        return key_name in src_arch.context

    def run_tests(self, detailed=False, silent=False):
        """
        Testcase schema
        testcase = {'title': kid[1].token_text(),
            'graph_ref': None, 'graph_block': None,
            'walker_ref': None, 'spawn_ctx': None,
            'assert_block': None, 'walker_block': None,
            'passed': None}
        """
        from pprint import pformat
        from time import time
        TY = '\033[33m'
        TG = '\033[32m'
        TR = '\033[31m'
        EC = '\033[m'
        num_failed = 0
        for i in self.testcases:
            destroy_set = []
            title = i['title']
            if(i['graph_ref']):
                gph = self.run_architype(
                    i['graph_ref'], kind='graph', caller=self)
            else:
                gph = architype(m_id=self._m_id, h=self._h,
                                parent_id=self.id,
                                code_ir=jac_ir_to_ast(i['graph_block']))
                destroy_set.append(gph)
                gph = gph.run()
            if(i['walker_ref']):
                wlk = self.spawn_walker(i['walker_ref'], caller=self)
            else:
                wlk = walker(m_id=self._m_id, h=self._h,
                             parent_id=self.id,
                             code_ir=jac_ir_to_ast(i['walker_block']))
                destroy_set.append(wlk)
            wlk.prime(gph)
            if(i['spawn_ctx']):
                self.run_spawn_ctx(jac_ir_to_ast(i['spawn_ctx']), wlk)
            stime = time()
            try:
                if(not silent):
                    print(f"Testing {title}: ", end='')
                wlk.run()
                if(i['assert_block']):
                    wlk._loop_ctrl = None
                    wlk.scope_and_run(jac_ir_to_ast(
                        i['assert_block']), run_func=wlk.run_code_block)
                i['passed'] = True
                if(not silent):
                    print(f"[{TG}PASSED{EC} in {TY}{time()-stime:.2f}s{EC}]")
                if(detailed and not silent):
                    print("REPORT: " + pformat(wlk.report))
            except Exception as e:
                i['passed'] = False
                num_failed += 1
                if(not silent):
                    print(f"[{TR}FAILED{EC} in {TY}{time()-stime:.2f}s{EC}]")
                    print(f"{e}")
            for i in destroy_set:  # FIXME: destroy set not complete
                i.destroy()
        num_tests = len(self.testcases)
        summary = {'tests': num_tests, 'passed': num_tests-num_failed,
                   'failed': num_failed,
                   'success': num_tests and not num_failed}
        if(detailed):
            details = []
            for i in self.testcases:
                details.append({'test': i['title'], 'passed': i['passed']})
            summary['details'] = details
        return summary

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.arch_ids.obj_list() + \
                self.walker_ids.obj_list():
            i.destroy()
        super().destroy()



from setuptools import setup, find_packages
setup(
    name='jaseci',
    version='1.3.2.11',
    packages=find_packages(include=['jaseci', 'jaseci.*']),
    install_requires=[
        'click>=8.1.0,<8.2.0',
        'click-shell>=2.1,<2.2',
        'numpy >= 1.22.3, < 1.23.0',
        'antlr4-python3-runtime>=4.9.3,<4.10.0',
        'fastapi[all]>=0.75.0,<1.0.0',
        'requests', 'redis', 'flake8', 'stripe',
        'pydantic'
    ],
    package_data={"": ["*.ini"], },
    entry_points={
        'console_scripts': [
            'jsctl = jaseci.jsctl.jsctl:main'
        ]
    })



import io
import pstats
import cProfile
import pdb
import uuid
import importlib
import pkgutil
import logging
import types
import base64
import json
import functools
import traceback
from time import time
from datetime import datetime
from pprint import pformat


# Get an instance of a logger
def connect_logger_handler(target_logger, handler, level=logging.WARN):
    """Attaches standard formatting and adds handler to logger"""
    target_logger.setLevel(level)
    handler.setFormatter(
        logging.Formatter(
            '%(asctime)s - %(levelname)s - %(funcName)s: %(message)s')
    )
    target_logger.addHandler(handler)


logger = logging.getLogger('core')
if(len(logger.handlers) < 1):
    connect_logger_handler(logger, logging.StreamHandler(), logging.INFO)

app_logger = logging.getLogger('app')
if(len(app_logger.handlers) < 1):
    connect_logger_handler(app_logger, logging.StreamHandler(), logging.INFO)


def log_var_out(val):
    """Print to log"""
    if(not logging.getLogger('core').disabled):
        logger.info(pformat(val))


def bp():
    pdb.set_trace()
    breakpoint()


def dummy_bp(inspect):
    traceback.print_stack()
    print(inspect)
    input()


def print_stack_to_log():
    tb = traceback.extract_stack()
    log_var_out(tb)


def is_urn(s: str):
    """Test if is uuid string in urn format"""
    return type(s) == str and s.startswith("urn:uuid:")


def is_jsonable(x):
    """Test if object can be json serialized"""
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


def json_out(val):
    if(type(val) == str):
        return val
    try:
        return json.dumps(val)
    except Exception:
        return val


def parse_str_token(s):
    return str(bytes(s, "utf-8").
               decode("unicode_escape")[1:-1])


def get_all_subclasses(cls):
    """Return list of all subclasses of cls"""
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_all_subclasses(c)])


def matching_fields(obj1, obj2):
    """
    Return list of matching member attributes in objects
    (non-private fields only)
    """
    matches = []
    for a in dir(obj1):
        if not a.startswith('_'):
            for b in dir(obj2):
                if a == b:
                    matches.append(a)
    return matches


def map_assignment_of_matching_fields(dest, source):
    """
    Assign the values of identical feild names from source to destination.
    """
    for i in matching_fields(dest, source):
        if (type(getattr(source, i)) == uuid.UUID):
            setattr(dest, i, getattr(source, i).urn)
        elif (type(getattr(source, i)) == datetime):
            setattr(dest, i, getattr(source, i).isoformat())
        elif not callable(getattr(dest, i)):
            setattr(dest, i, getattr(source, i))


obj_class_cache = {}


def build_class_dict(from_where):
    global obj_class_cache
    prefix = from_where.__name__ + "."
    for importer, modname, ispkg in \
            pkgutil.iter_modules(from_where.__path__, prefix):
        if(not ispkg):
            cls = modname.split('.')[-1]
            if(hasattr(importlib.import_module(modname), cls)):
                obj_class_cache[cls] = getattr(
                    importlib.import_module(modname), cls)
        else:
            if hasattr(from_where, modname.split('.')[-1]):
                build_class_dict(
                    getattr(from_where, modname.split('.')[-1])
                )


def find_class_and_import(class_name, from_where):
    """
    Search for class through all core packages

    Classes assumed to have same name as module file
    """
    global obj_class_cache
    if(not obj_class_cache.keys()):
        build_class_dict(from_where)
    return obj_class_cache[class_name]


def copy_func(f, name=None):
    """
    Utility to duplicate function in python.
    Can be used to programatically add methods to classes
    """
    g = types.FunctionType(f.__code__, f.__globals__,
                           name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    g.__name__ = name if name else g.__name__
    return g


def b64decode_str(code):
    """Decode a base 64 encoded string"""
    try:
        code = base64.b64decode(code).decode()
    except UnicodeDecodeError:
        logger.error(
            f'Code encoding invalid!')
    return code


class TestCaseHelper():
    """Helper to pretty print test results"""

    def setUp(self):
        self.logger_off()
        self.stime = time()
        return super().setUp()

    def tearDown(self):
        TY = '\033[33m'
        TG = '\033[32m'
        TR = '\033[31m'
        EC = '\033[m'
        td = super().tearDown()
        result = f'Time: {TY}{time()-self.stime:.3f} ' + \
                 f'- {EC}{self.id().split(".")[-1]}: '
        get_outcome = self.defaultTestResult()
        self._feedErrorsToResult(get_outcome, self._outcome.errors)
        if (len(get_outcome.errors) or len(get_outcome.failures)):
            result += f'{TR}[failed]{EC}'
        elif (len(self._outcome.skipped)):
            result += f'{TY}[skipped]{EC}'
        else:
            result += f'{TG}[passed]{EC}'

        print(result)
        self.logger_on()
        return td

    def logger_off(self):
        """Turn off logging output"""
        logging.getLogger('core').disabled = True
        logging.getLogger('app').disabled = True

    def logger_on(self):
        """Turn on logging output"""
        logging.getLogger('core').disabled = False
        logging.getLogger('app').disabled = False

    def is_logger_off(self):
        return (logging.getLogger('core').disabled and
                logging.getLogger('app').disabled)

    def log(self, val):
        """Print to log"""
        is_off = self.is_logger_off()
        self.logger_on()
        log_var_out(val)
        if(is_off):
            self.logger_off()

    def stopper(self):
        """Force test to fail"""
        self.assertTrue(False)

    def start_perf_test(self):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def stop_perf_test(self):
        self.pr.disable()
        s = io.StringIO()
        sortby = pstats.SortKey.CUMULATIVE
        ps = pstats.Stats(self.pr, stream=s).sort_stats(sortby)
        ps.print_stats(100)
        print(s.getvalue())


class bunch:
    def __init__(self, **kwds):
        self.__dict__.update(kwds)



"""
Variable scope manager for Jac

Utility for all runtime interaction with variables in different scopes
"""
from jaseci.utils.id_list import id_list
from jaseci.jac.machine.jac_value import jac_value
from jaseci.jac.machine.jac_value import jac_elem_wrap


class jac_scope():
    def __init__(self, parent, has_obj, action_sets):
        self.parent = parent
        self.local_scope = {}
        self.has_obj = has_obj if has_obj else self
        self.context = {}
        self.action_sets = action_sets + \
            [id_list(parent, in_list=parent._h.global_action_list)]
        self.setup_actions()

    def setup_actions(self):
        allactions = []
        for i in self.action_sets:
            allactions += i.obj_list()
        self.action_sets = {}
        for i in allactions:
            self.add_action(i)

    def add_action(self, act):
        group = act.name.split('.')[0]
        if '.' in act.name:
            if(group not in self.action_sets.keys()):
                self.action_sets[group] = {}
            action = act.name.split('.')[1]
            self.action_sets[group][action] = act
        else:
            self.action_sets[group] = act

    def set_agent_refs(self, cur_node, cur_walker):
        self.local_scope['here'] = jac_elem_wrap(cur_node)
        self.local_scope['visitor'] = jac_elem_wrap(cur_walker)

    def inherit_agent_refs(self, src_scope):
        self.local_scope['here'] = src_scope.local_scope['here']
        self.local_scope['visitor'] = src_scope.local_scope['visitor']

    def get_aganet_refs(self):
        return {
            'here': self.local_scope['here'],
            'visitor': self.local_scope['visitor']
        }

    def find_live_attr(self, name, allow_read_only=True):
        """Finds binding for variable if not in standard scope"""
        # check if var is in walker's context
        if('context' in self.has_obj.__dict__ and
           name in self.has_obj.context.keys()):
            return jac_value(self.parent,
                             ctx=self.has_obj,
                             name=name)
        elif(name in self.action_sets.keys()):
            return jac_value(self.parent,
                             ctx=self.action_sets,
                             name=name)
        return None

    def get_live_var(self, name, create_mode=False):
        """Returns live variable"""
        found = None
        # Lock for variable in various locations
        if (name in self.local_scope.keys()):
            found = jac_value(self.parent,
                              ctx=self.local_scope, name=name)
        else:
            found = self.find_live_attr(name)
        if (found is None and create_mode):
            self.local_scope[name] = None
            return jac_value(self.parent, ctx=self.local_scope, name=name)
        if(found):
            found.unwrap()
            return found
        return None



"""
Command line tool for Jaseci
"""

import click
from click_shell import shell
import os
import pickle
import functools
import json
import requests

from jaseci.utils.mem_hook import mem_hook
from jaseci.utils.utils import copy_func
from jaseci.element.super_master import super_master
from .book_tools import book

session = None
connection = None


def reset_state():
    global session, connection
    session = {
        "filename": "js.session",
        "user": [super_master(h=mem_hook(), name='admin')],
        "mem-only": session["mem-only"] if session is not None else False
    }
    session['master'] = session['user'][0]

    connection = {'url': None, 'token': None, 'headers': None}


reset_state()


@shell(prompt='jaseci > ', intro='Starting Jaseci Shell...')
@click.option('--filename', '-f', default="js.session",
              help="Specify filename for session state.")
@click.option('--mem-only', '-m', is_flag=True,
              help="Set true to not save file for session.")
def jsctl(filename, mem_only):
    """
    The Jaseci Command Line Interface
    """
    session['mem-only'] = mem_only
    session['filename'] = filename if not mem_only else None
    if (not mem_only and os.path.isfile(filename)):
        with open(filename, 'rb') as f:
            session['master'] = pickle.load(f)


def remote_api_call(payload, api_name):
    """
    Constructs and issues call to remote server
    NOTE: Untested
    """
    for i in super_master.all_apis(None):
        if(api_name == '_'.join(i['groups'])):
            if(i in super_master._private_api):
                path = '/jac/'+api_name
            elif(i in super_master._admin_api):
                path = '/admin/'+api_name
            elif(i in super_master._public_api):
                path = '/public/'+api_name
            break
    ret = requests.post(connection['url']+path,
                        json=payload,
                        headers=connection['headers'])
    if ret.status_code > 205:
        ret = f"Status Code Error {ret.status_code}"
    else:
        ret = ret.json()
    return ret


def resolve_none_type(kwargs):
    for i in kwargs.keys():
        if(kwargs[i] == 'None'):
            kwargs[i] = None


def interface_api(api_name, is_public, **kwargs):
    """
    Interfaces Master apis after processing arguments/parameters
    from cli
    """
    if('code' in kwargs and kwargs['code']):
        if (os.path.isfile(kwargs['code'])):
            with open(kwargs['code'], 'r') as file:
                if(api_name == 'sentinel_register' and
                   'name'in kwargs and kwargs['name'] == 'default'):
                    kwargs['name'] = kwargs['code']
                kwargs['code'] = file.read()
        else:
            click.echo(f"Code file {kwargs['code']} not found!")
            return
    if('ctx' in kwargs):  # can replace these checs with a dict check
        kwargs['ctx'] = json.loads(kwargs['ctx'])
    if('other_fields' in kwargs):
        kwargs['other_fields'] = json.loads(kwargs['other_fields'])
    resolve_none_type(kwargs)
    if(connection['token'] and connection['url']):
        out = remote_api_call(kwargs, api_name)
    elif(is_public):
        out = session['master'].public_interface_to_api(kwargs, api_name)
    else:
        out = session['master'].general_interface_to_api(kwargs, api_name)
    if(isinstance(out, dict) or isinstance(out, list)):
        out = json.dumps(out, indent=2)
    click.echo(out)
    if('output' in kwargs and kwargs['output']):
        with open(kwargs['output'], 'w') as f:
            f.write(out)
        click.echo(f'[saved to {kwargs["output"]}]')
    if not session['mem-only']:
        with open(session['filename'], 'wb') as f:
            pickle.dump(session['master'], f)


def extract_api_tree():
    """
    Generates a tree of command group names and function
    signatures in leaves from API function names in Master
    """
    api_funcs = {}
    for i in session['master'].all_apis() + \
            session['master']._cli_api:
        # Build hierarchy of command groups
        api_root = api_funcs
        for j in i['groups']:
            if (j not in api_root.keys()):
                api_root[j] = {}
            api_root = api_root[j]
        api_root['leaf'] = ['_'.join(i['groups']), i['sig'],
                            i in session['master']._public_api,  i['doc'],
                            i['cli_args']]
    return api_funcs


def build_cmd(group_func, func_name, leaf):
    """
    Generates Click function with options for each command
    group and leaf signatures
    leaf is format: [api_name, func_sig, is_public, func_doc]
    """

    f = functools.partial(
        copy_func(interface_api, func_name),
        api_name=leaf[0], is_public=leaf[2])
    f.__name__ = func_name
    f.__doc__ = leaf[3]

    func_sig = leaf[1]
    cli_args = leaf[4]
    for i in func_sig.parameters.keys():
        if(i == 'self'):
            continue
        p_default = func_sig.parameters[i].default
        p_type = func_sig.parameters[i].annotation
        if(p_type not in [int, bool, float]):
            p_type = str
        if(i in cli_args):
            f = click.argument(f'{i}', type=p_type)(f)
        elif(p_default is not func_sig.parameters[i].empty):
            f = click.option(
                f'-{i}', default=p_type(p_default), required=False,
                type=p_type)(f)
        else:
            f = click.option(
                f'-{i}', required=True, type=p_type)(f)
    # to file option to dump response to a file
    f = click.option(
        '--output', '-o', default='', required=False, type=str,
        help="Filename to dump output of this command call.")(f)
    return group_func.command()(f)


def cmd_tree_builder(location, group_func=jsctl, cmd_str=''):
    """
    Generates Click command groups from API tree recursively
    """
    for i in location.keys():
        loc = location[i]
        if ('leaf' in loc):
            build_cmd(group_func, i, loc['leaf'])
            continue
        else:
            f = copy_func(lambda: None, i)
            f.__doc__ = f'Group of `{(cmd_str+" "+i).lstrip()}` commands'
            new_func = group_func.group()(f)
        cmd_tree_builder(loc, new_func, cmd_str+' '+i)


@click.command(help="Command to log into live Jaseci server")
@click.argument("url", type=str, required=True)
@click.option('--username', '-u', required=True, prompt=True,
              help="Username to be used for login.")
@click.password_option(help="Password to be used for login.",
                       confirmation_prompt=False)
def login(url, username, password):
    url = url[:-1] if url[-1] == '/' else url
    payload = {'email': username, 'password': password}
    r = requests.post(url+'/user/token/', data=payload).json()
    if('token' in r.keys()):
        connection['token'] = r['token']
        connection['url'] = url
        connection['headers'] = {
            'Authorization': 'token ' + r['token']}
        click.echo("Login successful!")
    else:
        click.echo(f"Login failed!\n{r}")


@click.command(help="Edit a file")
@click.argument("file", type=str, required=True)
def edit(file):
    click.edit(filename=file)


@click.command(help="List relevant files")
@click.option('--all', '-a', is_flag=True,
              help="Flag for listing all files, not just relevant files")
def ls(all):
    for i in os.listdir():
        if(all):
            click.echo(f"{i}")
        else:
            click.echo(f"{i}") if i.endswith(
                '.jac') or i.endswith('.dot') else False


@click.command(help="Clear terminal")
def clear():
    click.clear()


@click.command(help="Reset jsctl (clears state)")
def reset():
    reset_state()
    click.echo(f"Jaseci State Cleared!")


@click.command(help="Internal book generation tools")
@click.argument("op", type=str, default='cheatsheet', required=True)
@click.option('--output', '-o', default='', required=False, type=str,
              help="Filename to dump output of this command call.")
def tool(op, output):
    out = ''
    if(op == 'cheatsheet'):
        out = f"{book().api_cheatsheet(extract_api_tree())}"
    elif(op == 'classes'):
        out = book().api_spec()
    click.echo(out)
    if(output):
        with open(output, 'w') as f:
            f.write(out)
        click.echo(f'[saved to {output}]')


jsctl.add_command(login)
jsctl.add_command(edit)
jsctl.add_command(ls)
jsctl.add_command(clear)
jsctl.add_command(reset)
jsctl.add_command(tool)
cmd_tree_builder(extract_api_tree())


def main():
    jsctl()


if __name__ == '__main__':
    main()



"""Built in actions for Jaseci"""
from jaseci.actions.live_actions import jaseci_action
from jaseci.jac.jac_set import jac_set


@jaseci_action()
def max(item_set: jac_set):
    ret = None
    if (not len(item_set)):
        return None
    items = item_set.obj_list()
    max_val = items[0].anchor_value()
    ret = items[0]
    for i in items:
        if(i.anchor_value() > max_val):
            ret = i
            max_val = i.anchor_value()
    return ret
